"""Derived from https://github.com/mfkasim1/xcnn/blob/f2cb9777da2961ac553f256ecdcca3e314a538ca/xcdnn2/litmodule.py"""
from deepchem.models.dft.scf import XCNNSCF
import torch
from deepchem.models.dft.nnxc import HybridXC
from deepchem.models.losses import Loss, L2Loss
from deepchem.models.torch_models.torch_model import TorchModel
from typing import Tuple, Optional, List, Any
import numpy as np


class DFTXC(torch.nn.Module):
    """
    This layer initializes the neural network exchange correlation functional and
    the hybrid functional. It is then used to run the Kohn Sham iterations.

    Examples
    --------
    >>> import torch
    >>> from deepchem.feat.dft_data import DFTEntry
    >>> from deepchem.models.dft.dftxc import DFTXC
    >>> e_type = 'ie'
    >>> true_val= '0.53411947056'
    >>> systems = [{'moldesc': 'N 0 0 0',
    ...       'basis': '6-311++G(3df,3pd)',
    ...        'spin': '3'},
    ...       {'moldesc': 'N 0 0 0',
    ...       'basis': '6-311++G(3df,3pd)',
    ...       'charge': 1,
    ...        'spin': '2'}]
    >>> entry = DFTEntry.create(e_type, true_val, systems)
    >>> nnmodel = _construct_nn_model(input_size=2, hidden_size=10, n_layers=1,modeltype=1).to(torch.double)
    >>> model = DFTXC("lda_x", nnmodel)
    >>> output = model([entry])
    The 6-311++G(3df,3pd) basis for atomz 7 does not exist, but we will download it
    Downloaded to /usr/share/miniconda3/envs/deepchem/lib/python3.8/site-packages/dqc/api/.database/6-311ppg_3df_3pd_/07.gaussian94

    """

    def __init__(self, xcstr: str, nnmodel: torch.nn.Module):
        """
        Parameters
        ----------
        xcstr: str
            The choice of xc to use. Some of the commonly used ones are:
            lda_x, lda_c_pw, lda_c_ow, lda_c_pz, lda_xc_lp_a, lda_xc_lp_b.
        nnmodel: torch.nn.Module
            the PyTorch model implementing the calculation

        Notes
        -----
        It is not necessary to use the default method(_construct_nn_model) with the XCModel.
        """
        super(DFTXC, self).__init__()
        self.xcstr = xcstr
        self.nnmodel = nnmodel

    def forward(self, inputs):
        """
        Parameters
        ----------
        inputs: list
            list of entry objects that have been defined using DFTEntry

        Returns
        -------
        output: list of torch.Tensor
            Calculated value of the data point after running the Kohn Sham iterations
            using the neural network XC functional.
        """
        hybridxc = HybridXC(self.xcstr, self.nnmodel, aweight0=0.0)
        output = []
        for entry in inputs:
            evl = XCNNSCF(hybridxc, entry)
            qcs = []
            for system in entry.get_systems():
                qcs.append(evl.run(system))
            if entry.entry_type == 'dm':
                output.append((torch.as_tensor(entry.get_val(qcs)[0])))
            else:
                output.append(
                    torch.tensor(entry.get_val(qcs), requires_grad=True))
        return output


class XCModel(TorchModel):
    """
    This class is used to initialize and run Differentiable Quantum Chemistry (i.e,
    DFT) calculations, using an exchange correlation functional that has been replaced
    by a neural network. This model is based on the paper "Learning the exchange-correlation
    functional from nature with fully differentiable density functional
    theory." and is listed below for reference.

    To read more about Density Functional Theory and the exchange
    correlation functional please check the references below.

    Examples
    --------
    >>> from deepchem.models.dft.dftxc import XCModel
    >>> from deepchem.data.data_loader import DFTYamlLoader
    >>> inputs = 'deepchem/models/tests/assets/test_dftxcdata.yaml'
    >>> data = DFTYamlLoader()
    >>> dataset = data.create_dataset(inputs)
    >>> dataset.get_shape()
    ((2,), (2,), (2,), (2,))
    >>> model = XCModel("lda_x", batch_size=1)
    >>> loss = model.fit(dataset, nb_epoch=1, checkpoint_interval=1)
    The 6-311++G(3df,3pd) basis for atomz 3 does not exist, but we will download it
    Downloaded to /usr/share/miniconda3/envs/deepchem/lib/python3.8/site-packages/dqc/api/.database/6-311ppg_3df_3pd_/03.gaussian94

    Notes
    -----
    There are 4 types of DFT data object implementations that are used to determine the type
    of calculation to be carried out on the entry object. These types are: "ae", "ie", "dm",    "dens", that stand for atomization energy, ionization energy, density matrix and
    density profile respectively.
    The entry type "Density Matrix" cannot be used on model.evaluate as of now.
    To run predictions on this data type, a dataset containing only "dm" entries must
    be used.

    References
    ----------
    https://github.com/deepchem/deepchem/blob/3f06168a6c9c16fd90cde7f5246b94f484ea3890/deepchem/models/dft/nnxc.py
    Encyclopedia of Condensed Matter Physics, 2005.
    Kasim, Muhammad F., and Sam M. Vinko. "Learning the exchange-correlation
    functional from nature with fully differentiable density functional
    theory." Physical Review Letters 127.12 (2021): 126403.
    """

    def __init__(self,
                 xcstr: str,
                 nnmodel: Optional[torch.nn.Module] = None,
                 input_size: int = 2,
                 hidden_size: int = 10,
                 n_layers: int = 1,
                 modeltype: int = 1,
                 n_tasks: int = 0,
                 log_frequency: int = 0,
                 mode: str = 'classification',
                 device: Optional[torch.device] = None,
                 **kwargs) -> None:
        """
        Parameters
        ----------
        xcstr: str
            The choice of xc to use.
        nnmodel: torch.nn.Module
            the PyTorch model implementing the calculation
        input_size: int
            size of neural network input
        hidden_size: int
            size of the hidden layers ; the number of hidden layers is fixed
            in the default method.
        n_layers: int
            number of layers in the neural network
        modeltype: int
            model type 2 includes an activation layer whereas type 1 does not.
        """
        if nnmodel is None:
            nnmodel = _construct_nn_model(input_size, hidden_size, n_layers,
                                          modeltype).to(torch.double)
        model = (DFTXC(xcstr, nnmodel)).to(device)
        self.xc = xcstr
        loss: Loss = L2Loss()
        output_types = ['loss', 'predict']
        self.mode = mode
        super(XCModel, self).__init__(model,
                                      loss=loss,
                                      output_types=output_types,
                                      **kwargs)

    def _prepare_batch(
            self,
            batch) -> Tuple[List[Any], List[torch.Tensor], List[torch.Tensor]]:
        """
        Method to compute inputs, labels and weight for the Torch Model.

        Parameters
        ----------
        batch: Tuple[Any, Any, Any]

        Returns
        ------
        Tuple[List[torch.Tensor], List[torch.Tensor], List[torch.Tensor]]
        """
        inputs, labels, weights = batch
        if labels is not None:
            labels = [
                x.astype(np.float32) if x.dtype == np.float64 else x
                for x in labels
            ]
            label_tensors = [
                torch.as_tensor(x, dtype=torch.float64,
                                device=self.device).requires_grad_()
                for x in labels
            ]
        else:
            label_tensors = []
        if weights is not None:
            weights = [
                x.astype(np.float32) if x.dtype == np.float64 else x
                for x in weights
            ]
            weight_tensors = [
                torch.as_tensor(x, dtype=torch.float64, device=self.device)
                for x in weights
            ]
        else:
            weight_tensors = []

        return (inputs, label_tensors, weight_tensors)


class ExpM1Activation(torch.nn.Module):
    """
    This class is an activation layer that is used with model_type 2.

    Examples
    --------
    >>> from deepchem.models.dft.dftxc import ExpM1Activation
    >>> import torch
    >>> model = ExpM1Activation()
    >>> x = torch.tensor(2.5)
    >>> output = model(x)
    """

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.exp(x) - 1


def _construct_nn_model(input_size: int, hidden_size: int, n_layers: int,
                        modeltype: int):
    """
    Constructs Neural Network

    Parameters
    ----------
    input_size: int
        size of neural network input
    hidden_size: int
        size of the hidden layers ; there are 3 hidden layers in this method
    n_layers: int
        number of layers in the neural network
    modeltype: int
        model type 2 includes an activation layer whereas type 1 does not.

    Returns
    -------
    torch.nn.Sequential(*layers)

    Notes
    -----
    It is not necessary to use this method with the XCModel, user defined pytorch
    models will work.
    """
    if modeltype == 1:
        layers: List[Any]
        layers = []
        for i in range(n_layers):
            n1 = input_size if i == 0 else hidden_size
            layers.append(torch.nn.Linear(n1, hidden_size))
            layers.append(torch.nn.Softplus())
        layers.append(torch.nn.Linear(hidden_size, 1, bias=False))
        return torch.nn.Sequential(*layers)
    elif modeltype == 2:
        layers = []
        for i in range(n_layers):
            n1 = input_size if i == 0 else hidden_size
            layers.append(torch.nn.Linear(n1, hidden_size))
            if i < n_layers - 1:
                layers.append(torch.nn.Softplus())
            else:
                layers.append(ExpM1Activation())
        layers.append(torch.nn.Linear(hidden_size, 1, bias=False))
        return torch.nn.Sequential(*layers)



import pytest
try:
    import torch
    from deepchem.utils.differentiation_utils import EditableModule
    has_torch = True
except ModuleNotFoundError:
    has_torch = False


@pytest.mark.torch
def test_getparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(0.4)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(1.6)
    assert alpha.getparams("mult") == [torch.tensor(4.)]


@pytest.mark.torch
def test_setparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(4.0)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(16)
    alpha.setparams("mult", torch.tensor(5.0))
    assert alpha.mult(x) == torch.tensor(20.0)


@pytest.mark.torch
def test_cached_getparamnames():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    alpha = A(a)
    assert alpha.cached_getparamnames("mult") == ["b"]


@pytest.mark.torch
def test_getuniqueparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b**2 * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(0.4)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(6.4)
    assert alpha.getuniqueparams("mult") == [torch.tensor(4.)]  # Not 16.0


@pytest.mark.torch
def test_setuniqueparams():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b**2 * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(0.4)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(6.4)
    assert alpha.getuniqueparams("mult") == [torch.tensor(4.)]
    alpha.setuniqueparams("mult", torch.tensor(5.0))
    assert alpha.mult(x) == torch.tensor(10.0)


@pytest.mark.torch
def test_get_unique_params_idxs():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a
            self.c = a * a * a

        def mult(self, x):
            return self.b * self.c * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b", prefix + "c"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(4.0)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(128.0)
    assert alpha.getparams("mult") == [torch.tensor(4.), torch.tensor(8.)]
    assert alpha._get_unique_params_idxs("mult") == [0, 1]


@pytest.mark.torch
def test_assertparams():
    """Test that assertparams works correctly.
    also checks the private methods as they are used in it.
    - __assert_method_preserve
    - __assert_get_correct_params
    - __list_operating_params

    """

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a
            self.c = a * a * a

        def mult(self, x):
            return self.b * self.c * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b", prefix + "c"]
            else:
                raise KeyError()

    a = torch.tensor(2.0)
    x = torch.tensor(4.0)
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(128.0)
    assert alpha.getparams("mult") == [torch.tensor(4.), torch.tensor(8.)]
    alpha.assertparams(alpha.mult, x)


@pytest.mark.torch
def test_getparamnames():

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    a = torch.tensor(2.0).requires_grad_()
    x = torch.tensor(0.4).requires_grad_()
    alpha = A(a)
    assert alpha.mult(x) == torch.tensor(1.6000)
    assert alpha.getparamnames("mult") == ['b']


@pytest.mark.torch
def test_traverse_obj():
    from deepchem.utils.differentiation_utils.editable_module import _traverse_obj, torch_float_type

    class A:

        def __init__(self):
            self.a = 2
            self.b = torch.tensor(3.0)
            self.c = torch.tensor(4.0)
            self.d = torch.tensor(5.0)

    a = A()

    def action(elmt, name, objdict, key):
        print(name, elmt)

    def crit(elmt):
        return isinstance(elmt, torch.Tensor) and elmt.dtype in torch_float_type

    a = _traverse_obj(a, "", action, crit)  # Check Doesn't Crashes


@pytest.mark.torch
def test_get_tensor():
    from deepchem.utils.differentiation_utils.editable_module import _get_tensors

    class A:

        def __init__(self):
            self.a = 2
            self.b = torch.tensor(3.0)
            self.c = torch.tensor(4.0)
            self.d = torch.tensor(5.0)

    a = A()
    outputs = _get_tensors(a)
    assert outputs[0] == [torch.tensor(3.), torch.tensor(4.), torch.tensor(5.)]
    assert outputs[1] == ['b', 'c', 'd']


@pytest.mark.torch
def test_set_tensor():
    from deepchem.utils.differentiation_utils.editable_module import _set_tensors

    class A:

        def __init__(self):
            self.a = 2
            self.b = torch.tensor(3.0)
            self.c = torch.tensor(4.0)
            self.d = torch.tensor(5.0)

    a = A()
    _set_tensors(a, [torch.tensor(6.), torch.tensor(7.), torch.tensor(8.)])
    assert a.b == torch.tensor(6.)
    assert a.c == torch.tensor(7.)


@pytest.mark.torch
def test_normalize_bcast_dims():
    from deepchem.utils.differentiation_utils import normalize_bcast_dims
    assert normalize_bcast_dims([1, 2, 3], [2, 3]) == [[1, 2, 3], [1, 2, 3]]


@pytest.mark.torch
def test_get_bcasted_dims():
    from deepchem.utils.differentiation_utils import get_bcasted_dims
    assert get_bcasted_dims([1, 2, 5], [2, 3, 4]) == [2, 3, 5]


@pytest.mark.torch
def test_match_dim():
    from deepchem.utils.differentiation_utils import match_dim
    x = torch.randn(10, 5)
    xq = torch.randn(10, 3)
    x_new, xq_new = match_dim(x, xq)
    assert x_new.shape == torch.Size([10, 5])
    assert xq_new.shape == torch.Size([10, 3])


@pytest.mark.torch
def test_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator
    torch.manual_seed(100)

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

        def _rmv(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _mm(self, x):
            return torch.matmul(self.param, x)

        def _rmm(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _fullmatrix(self):
            return self.param

    linop = MyLinOp((1, 3, 1, 2))
    x = torch.rand(1, 3, 2, 2)
    assert torch.allclose(linop.mv(x), torch.matmul(linop.param, x))
    x = torch.rand(1, 3, 1, 1)
    assert torch.allclose(linop.rmv(x),
                          torch.matmul(linop.param.transpose(-2, -1).conj(), x))
    x = torch.rand(1, 3, 2, 2)
    assert torch.allclose(linop.mm(x), torch.matmul(linop.param, x))
    x = torch.rand(1, 3, 1, 2)
    assert torch.allclose(linop.rmm(x),
                          torch.matmul(linop.param.transpose(-2, -1).conj(), x))
    assert torch.allclose(linop.fullmatrix(), linop.param)


@pytest.mark.torch
def test_add_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class Operator(LinearOperator):

        def __init__(self, mat: torch.Tensor, is_hermitian: bool) -> None:
            super(Operator, self).__init__(
                shape=mat.shape,
                is_hermitian=is_hermitian,
                dtype=mat.dtype,
                device=mat.device,
                _suppress_hermit_warning=True,
            )
            self.mat = mat

        def _mv(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(self.mat, x.unsqueeze(-1)).squeeze(-1)

        def _mm(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(self.mat, x)

        def _rmv(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(
                self.mat.transpose(-3, -1).conj(), x.unsqueeze(-1)).squeeze(-1)

        def _rmm(self, x: torch.Tensor) -> torch.Tensor:
            return torch.matmul(self.mat.transpose(-2, -1).conj(), x)

        def _fullmatrix(self) -> torch.Tensor:
            return self.mat

        def _getparamnames(self, prefix: str = ""):
            return [prefix + "mat"]

    op = Operator(torch.tensor([[1, 2.], [3, 4]]), is_hermitian=False)
    x = torch.tensor([[2, 2], [1, 2.]])
    op2 = op + op
    assert torch.allclose(op2.mm(x), 2 * op.mm(x))


@pytest.mark.torch
def test_mul_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

    linop = MyLinOp((1, 3, 1, 2))
    linop2 = linop * 2
    x = torch.rand(1, 3, 2, 2)
    torch.allclose(linop.mv(x) * 2, linop2.mv(x))


@pytest.mark.torch
def test_adjoint_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

        def _rmv(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

    linop = MyLinOp((1, 3, 1, 2))
    x = torch.rand(1, 3, 1, 1)
    result_rmv = linop.rmv(x)

    adjoint_linop = linop.H
    result_mv = adjoint_linop.mv(x)

    assert torch.allclose(result_rmv, result_mv)


@pytest.mark.torch
def test_matmul_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

    linop1 = MyLinOp((1, 3, 1, 2))
    linop2 = MyLinOp((1, 3, 2, 1))
    linop_result = linop1.matmul(linop2)
    x = torch.rand(1, 3, 1, 1)
    result = linop_result.mv(x)
    assert result.shape == torch.Size([1, 3, 1, 1])


@pytest.mark.torch
def test_matrix_linear_operator():
    from deepchem.utils.differentiation_utils import LinearOperator

    mat = torch.rand(2, 2)
    linop = LinearOperator.m(mat)
    x = torch.randn(2, 2)

    result_mm = linop.mm(x)
    expected_mm = torch.matmul(mat, x)

    result_mv = linop.mv(x)
    expected_mv = torch.matmul(mat, x.unsqueeze(-1)).squeeze(-1)

    assert torch.allclose(result_mm, expected_mm)
    assert torch.allclose(result_mv, expected_mv)


def test_set_default_options():
    from deepchem.utils.differentiation_utils import set_default_option
    assert set_default_option({'a': 1, 'b': 2}, {'a': 3}) == {'a': 3, 'b': 2}


def test_get_and_pop_keys():
    from deepchem.utils.differentiation_utils import get_and_pop_keys
    assert get_and_pop_keys({'a': 1, 'b': 2}, ['a']) == {'a': 1}


def test_get_method():
    from deepchem.utils.differentiation_utils import get_method
    assert get_method('foo', {'bar': lambda: 1}, 'bar')() == 1


def test_dummy_context_manager():
    """Just checks that dummy_context_manager doesn't crash"""
    from deepchem.utils.differentiation_utils import dummy_context_manager
    with dummy_context_manager() as x:
        if x is None:
            pass
        else:
            raise AssertionError()


def test_assert_runtime():
    from deepchem.utils.differentiation_utils import assert_runtime
    try:
        assert_runtime(False, "This should fail")
    except RuntimeError:
        pass


@pytest.mark.torch
def test_take_eigpairs():
    from deepchem.utils.differentiation_utils.symeig import _take_eigpairs
    eival = torch.tensor([[1., 2., 3.], [4., 5., 6.]])
    eivec = torch.tensor([[[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]],
                          [[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]]])
    neig = 2
    mode = "lowest"
    eival, eivec = _take_eigpairs(eival, eivec, neig, mode)
    assert torch.allclose(eival, torch.tensor([[1., 2.], [4., 5.]]))
    assert torch.allclose(
        eivec,
        torch.tensor([[[1., 2.], [4., 5.], [7., 8.]],
                      [[1., 2.], [4., 5.], [7., 8.]]]))


@pytest.mark.torch
def test_exacteig():
    from deepchem.utils.differentiation_utils.symeig import exacteig
    from deepchem.utils.differentiation_utils import LinearOperator
    torch.manual_seed(100)
    mat = LinearOperator.m(torch.randn(2, 2))
    eival, eivec = exacteig(mat, 2, "lowest", None)
    assert eival.shape == torch.Size([2])
    assert eivec.shape == torch.Size([2, 2])


@pytest.mark.torch
def test_degen_symeig():
    from deepchem.utils.differentiation_utils.symeig import degen_symeig
    from deepchem.utils.differentiation_utils import LinearOperator
    A = LinearOperator.m(torch.rand(2, 2))
    evals, evecs = degen_symeig.apply(A.fullmatrix())
    assert evals.shape == torch.Size([2])
    assert evecs.shape == torch.Size([2, 2])


@pytest.mark.torch
def test_davidson():
    from deepchem.utils.differentiation_utils.symeig import davidson
    from deepchem.utils.differentiation_utils import LinearOperator
    A = LinearOperator.m(torch.rand(2, 2))
    neig = 2
    mode = "lowest"
    eigen_val, eigen_vec = davidson(A, neig, mode)
    assert eigen_val.shape == torch.Size([2])
    assert eigen_vec.shape == torch.Size([2, 2])


@pytest.mark.torch
def test_pure_function():
    from deepchem.utils.differentiation_utils import PureFunction

    class WrapperFunction(PureFunction):

        def _get_all_obj_params_init(self):
            return []

        def _set_all_obj_params(self, objparams):
            pass

    def fcn(x, y):
        return x + y

    pfunc = WrapperFunction(fcn)
    assert pfunc(1, 2) == 3


@pytest.mark.torch
def test_function_pure_function():
    from deepchem.utils.differentiation_utils.pure_function import FunctionPureFunction

    def fcn(x, y):
        return x + y

    pfunc = FunctionPureFunction(fcn)
    assert pfunc(1, 2) == 3


@pytest.mark.torch
def test_editable_module_pure_function():
    from deepchem.utils.differentiation_utils import EditableModule
    from deepchem.utils.differentiation_utils.pure_function import EditableModulePureFunction

    class A(EditableModule):

        def __init__(self, a):
            self.b = a * a

        def mult(self, x):
            return self.b * x

        def getparamnames(self, methodname, prefix=""):
            if methodname == "mult":
                return [prefix + "b"]
            else:
                raise KeyError()

    B = A(4)
    m = EditableModulePureFunction(B, B.mult)
    m.set_objparams([3])
    assert m(2) == 6


@pytest.mark.torch
def test_torch_nn_pure_function():
    from deepchem.utils.differentiation_utils import get_pure_function

    class A(torch.nn.Module):

        def __init__(self, a):
            super().__init__()
            self.b = torch.nn.Parameter(torch.tensor(a * a))

        def forward(self, x):
            return self.b * x

    B = A(4.)
    m = get_pure_function(B.forward)
    m.set_objparams([3.])
    assert m(2) == 6.0


@pytest.mark.torch
def test_check_identical_objs():
    from deepchem.utils.differentiation_utils.pure_function import _check_identical_objs
    a = [1, 2, 3]
    assert _check_identical_objs([a], [a])


@pytest.mark.torch
def test_get_pure_function():
    from deepchem.utils.differentiation_utils import get_pure_function

    def fcn(x, y):
        return x + y

    pfunc = get_pure_function(fcn)
    assert pfunc(1, 2) == 3


@pytest.mark.torch
def test_make_siblings():
    from deepchem.utils.differentiation_utils import make_sibling

    def fcn1(x, y):
        return x + y

    @make_sibling(fcn1)
    def fcn3(x, y):
        return x * y

    assert fcn3(1, 2) == 2


@pytest.mark.torch
def test_wrap_gmres():
    from deepchem.utils.differentiation_utils.solve import wrap_gmres
    from deepchem.utils.differentiation_utils import LinearOperator
    A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    B = torch.tensor([[[5., 6], [7, 8]]])
    assert torch.allclose(A.fullmatrix() @ wrap_gmres(A, B, None, None), B)


@pytest.mark.torch
def test_exact_solve():
    from deepchem.utils.differentiation_utils.solve import exactsolve
    from deepchem.utils.differentiation_utils import LinearOperator
    A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    B = torch.tensor([[5., 6], [7, 8]])
    assert torch.allclose(A.fullmatrix() @ exactsolve(A, B, None, None), B)


@pytest.mark.torch
def test_solve_ABE():
    from deepchem.utils.differentiation_utils.solve import solve_ABE
    A = torch.tensor([[1., 2], [3, 4]])
    B = torch.tensor([[5., 6], [7, 8]])
    E = torch.tensor([1., 2])
    expected_result = torch.tensor([[-0.1667, 0.5000], [2.5000, 3.2500]])
    assert torch.allclose(solve_ABE(A, B, E), expected_result, 0.001)


@pytest.mark.torch
def test_get_batch_dims():
    from deepchem.utils.differentiation_utils.solve import get_batchdims
    from deepchem.utils.differentiation_utils import MatrixLinearOperator
    A = MatrixLinearOperator(torch.randn(4, 3, 3), True)
    B = torch.randn(3, 3, 2)
    assert get_batchdims(A, B, None,
                         None) == [max(A.shape[:-2], B.shape[:-2])[0]]


@pytest.mark.torch
def test_setup_precond():
    from deepchem.utils.differentiation_utils.solve import setup_precond
    from deepchem.utils.differentiation_utils import MatrixLinearOperator
    A = MatrixLinearOperator(torch.randn(4, 3, 3), True)
    B = torch.randn(4, 3, 2)
    cond = setup_precond(A)
    assert cond(B).shape == torch.Size([4, 3, 2])


@pytest.mark.torch
def test_dot():
    from deepchem.utils.differentiation_utils.solve import dot
    r = torch.tensor([[1, 2], [3, 4]])
    z = torch.tensor([[5, 6], [7, 8]])
    assert torch.allclose(dot(r, z), torch.tensor([[26, 44]]))
    assert torch.allclose(dot(r, z), sum(r * z))


@pytest.mark.torch
def test_gmres():
    from deepchem.utils.differentiation_utils.solve import gmres
    from deepchem.utils.differentiation_utils import MatrixLinearOperator
    A = MatrixLinearOperator(torch.tensor([[1., 2], [3, 4]]), True)
    B = torch.tensor([[5., 6], [7, 8]])
    expected_result = torch.tensor([[0.8959, 1.0697], [1.2543, 1.4263]])
    assert torch.allclose(gmres(A, B), expected_result, 0.001)


@pytest.mark.torch
def test_setup_linear_problem():
    from deepchem.utils.differentiation_utils import MatrixLinearOperator
    from deepchem.utils.differentiation_utils.solve import setup_linear_problem
    A = MatrixLinearOperator(torch.randn(4, 3, 3), True)
    B = torch.randn(4, 3, 2)
    A_fcn, AT_fcn, B_new, col_swapped = setup_linear_problem(
        A, B, None, None, [4], None, False)
    assert A_fcn(B).shape == torch.Size([4, 3, 2])


@pytest.mark.torch
def test_safe_denom():
    from deepchem.utils.differentiation_utils.solve import safedenom
    r = torch.tensor([[0., 2], [3, 4]])
    assert torch.allclose(
        safedenom(r, 1e-9),
        torch.tensor([[1.0000e-09, 2.0000e+00], [3.0000e+00, 4.0000e+00]]))


@pytest.mark.torch
def test_get_largest_eival():
    from deepchem.utils.differentiation_utils.solve import get_largest_eival

    def Afcn(x):
        return 10 * x

    x = torch.tensor([[1., 2], [3, 4]])
    assert torch.allclose(get_largest_eival(Afcn, x), torch.tensor([[10.,
                                                                     10.]]))


@pytest.mark.torch
def test_symeig():
    from deepchem.utils.differentiation_utils import LinearOperator, symeig
    A = LinearOperator.m(torch.tensor([[3, -1j], [1j, 4]]))
    evals, evecs = symeig(A)
    assert evecs.shape == torch.Size([2, 2])
    assert evals.shape == torch.Size([2])


@pytest.mark.torch
def test_check_degen():
    from deepchem.utils.differentiation_utils.symeig import _check_degen
    evals = torch.tensor([1, 1, 2, 3, 3, 3, 4, 5, 5])
    degen_atol = 0.1
    degen_rtol = 0.1
    idx_degen, isdegenerate = _check_degen(evals, degen_atol, degen_rtol)
    assert idx_degen.shape == torch.Size([9, 9])


@pytest.mark.torch
def test_ortho():
    from deepchem.utils.differentiation_utils import ortho
    A = torch.tensor([[1, 2], [3, 4]])
    B = torch.tensor([[1, 0], [0, 1]])
    assert torch.allclose(ortho(A, B), torch.tensor([[0, 2], [3, 0]]))


@pytest.mark.torch
def test_jac():
    from deepchem.utils.differentiation_utils import jac

    def fcn(x, y):
        return x * y

    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    n_lin = jac(fcn, [x, y])
    input = torch.tensor([[1, 3, 3], [4, 5, 6]])
    assert torch.allclose(n_lin[1].mv(input), x * input)


@pytest.mark.torch
def test_jac_class():
    from deepchem.utils.differentiation_utils.grad import get_pure_function, _Jac

    def fcn(x, y):
        return x * y

    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    pfcn = get_pure_function(fcn)
    assert torch.allclose(
        _Jac(pfcn, [x, y], 0).mv(torch.tensor([1.0, 1.0, 1.0])), y)


@pytest.mark.torch
def test_connect_graph():
    from deepchem.utils.differentiation_utils.grad import connect_graph
    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    out = x * y
    assert torch.allclose(connect_graph(out, [x, y]),
                          torch.tensor([4., 10., 18.]))


@pytest.mark.torch
def test_setup_idxs():
    from deepchem.utils.differentiation_utils.grad import _setup_idxs
    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = torch.tensor([4.0, 5.0, 6.0], requires_grad=True)
    assert _setup_idxs(None, [x, y]) == [0, 1]


@pytest.mark.torch
def test_solve():
    from deepchem.utils.differentiation_utils import LinearOperator, solve
    A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    B = torch.tensor([[5., 6], [7, 8]])
    assert torch.allclose(solve(A, B),
                          torch.tensor([[-3.0000, -4.0000], [4.0000, 5.0000]]))


@pytest.mark.torch
def test_cg():
    from deepchem.utils.differentiation_utils import LinearOperator, cg
    A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    B = torch.tensor([[5., 6], [7, 8]])
    assert torch.allclose(cg(A, B),
                          torch.tensor([[-3.0000, -4.0000], [4.0000, 5.0000]]))



import numpy as np

from deepchem.utils.typing import RDKitMol
from deepchem.feat.base_classes import MolecularFeaturizer


class MACCSKeysFingerprint(MolecularFeaturizer):
    """MACCS Keys Fingerprint.

    The MACCS (Molecular ACCess System) keys are one of the most commonly used structural keys.
    Please confirm the details in [1]_, [2]_.

    Examples
    --------
    >>> import deepchem as dc
    >>> smiles = 'CC(=O)OC1=CC=CC=C1C(=O)O'
    >>> featurizer = dc.feat.MACCSKeysFingerprint()
    >>> features = featurizer.featurize([smiles])
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (167,)

    References
    ----------
    .. [1] Durant, Joseph L., et al. "Reoptimization of MDL keys for use in drug discovery."
             Journal of chemical information and computer sciences 42.6 (2002): 1273-1280.
    .. [2] https://github.com/rdkit/rdkit/blob/master/rdkit/Chem/MACCSkeys.py

    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self):
        """Initialize this featurizer."""
        self.calculator = None

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """
        Calculate MACCS keys fingerprint.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
                RDKit Mol object

        Returns
        -------
        np.ndarray
                1D array of RDKit descriptors for `mol`. The length is 167.

        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )

        if self.calculator is None:
            try:
                from rdkit.Chem.AllChem import GetMACCSKeysFingerprint
                self.calculator = GetMACCSKeysFingerprint
            except ModuleNotFoundError:
                raise ImportError("This class requires RDKit to be installed.")

        return self.calculator(datapoint)



import torch
import inspect
from typing import Callable, List, Tuple, Union, Sequence
from deepchem.utils.attribute_utils import set_attr, del_attr
from deepchem.utils.differentiation_utils import EditableModule
from deepchem.utils.misc_utils import Uniquifier
from contextlib import contextmanager
from abc import abstractmethod


class PureFunction(object):
    """
    PureFunction class wraps methods to make it stateless and expose the pure
    function to take inputs of the original inputs (`params`) and the object's
    states (`objparams`).
    For functions, this class only acts as a thin wrapper.

    Restore stack stores list of (objparams, identical) everytime the objparams
    are set, it will store the old objparams and indication if the old and new
    objparams are identical.

    For Using this Class we first need to implement `_get_all_obj_params_init`
    and `_set_all_obj_params`.

    Examples
    --------
    >>> class WrapperFunction(PureFunction):
    ...     def _get_all_obj_params_init(self):
    ...         return []
    ...     def _set_all_obj_params(self, objparams):
    ...         pass
    >>> def fcn(x, y):
    ...    return x + y
    >>> pfunc = WrapperFunction(fcn)
    >>> pfunc(1, 2)
    3

    """

    def __init__(self, fcntocall: Callable):
        """Initialize the PureFunction.

        Parameters
        ----------
        fcntocall: Callable
            The function to be wrapped

        """
        self._state_change_allowed = True
        self._allobjparams = self._get_all_obj_params_init()
        self._uniq = Uniquifier(self._allobjparams)
        self._cur_objparams = self._uniq.get_unique_objs()
        self._fcntocall = fcntocall
        self._restore_stack: List[Tuple[List, bool]] = []

    def __call__(self, *params):
        """Call the wrapped function with the current object parameters and
        the input parameters.

        Parameters
        ----------
        params: tuple
            The input parameters of the wrapped function

        Returns
        -------
        Any
            The output of the wrapped function

        """
        return self._fcntocall(*params)

    @abstractmethod
    def _get_all_obj_params_init(self) -> List:
        """Get the initial object parameters.

        Returns
        -------
        List
            The initial object parameters

        """
        pass

    @abstractmethod
    def _set_all_obj_params(self, allobjparams: List):
        """Set the object parameters.

        Parameters
        ----------
        allobjparams: List
            The object parameters to be set

        """
        pass

    def objparams(self) -> List:
        """Get the current object parameters.

        Returns
        -------
        List
            The current object parameters

        """
        return self._cur_objparams

    def set_objparams(self, objparams: List):
        """Set the object parameters.

        Parameters
        ----------
        objparams: List
            The object parameters to be set

        TODO: check if identical with current object parameters

        """
        identical = _check_identical_objs(objparams, self._cur_objparams)
        self._restore_stack.append((self._cur_objparams, identical))
        if not identical:
            allobjparams = self._uniq.map_unique_objs(objparams)
            self._set_all_obj_params(allobjparams)
            self._cur_objparams = list(objparams)

    def restore_objparams(self):
        """Restore the object parameters to the previous state."""
        old_objparams, identical = self._restore_stack.pop(-1)
        if not identical:
            allobjparams = self._uniq.map_unique_objs(old_objparams)
            self._set_all_obj_params(allobjparams)
            self._cur_objparams = old_objparams

    @contextmanager
    def useobjparams(self, objparams: List):
        """Context manager to temporarily set the object parameters.

        Parameters
        ----------
        objparams: List
            The object parameters to be set temporarily

        """
        if not self._state_change_allowed:
            raise RuntimeError("The state change is disabled")
        try:
            self.set_objparams(objparams)
            yield
        finally:
            self.restore_objparams()

    @contextmanager
    def disable_state_change(self):
        """Context manager to temporarily disable the state change."""
        try:
            prev_status = self._state_change_allowed
            self._state_change_allowed = False
            yield
        finally:
            self._state_change_allowed = prev_status


class FunctionPureFunction(PureFunction):
    """Implementation of PureFunction for functions.
    It just acts as a thin wrapper for the function.

    Examples
    --------
    >>> def fcn(x, y):
    ...     return x + y
    >>> pfunc = FunctionPureFunction(fcn)
    >>> pfunc(1, 2)
    3

    """

    def _get_all_obj_params_init(self) -> List:
        """Get the initial object parameters.

        Returns
        -------
        List
            The initial object parameters

        """
        return []

    def _set_all_obj_params(self, objparams: List):
        """Set the object parameters.

        Parameters
        ----------
        objparams: List
            The object parameters to be set

        """
        pass


class EditableModulePureFunction(PureFunction):
    """Implementation of PureFunction for EditableModule.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import EditableModule, get_pure_function
    >>> class A(EditableModule):
    ...     def __init__(self, a):
    ...         self.b = a*a
    ...     def mult(self, x):
    ...         return self.b * x
    ...     def getparamnames(self, methodname, prefix=""):
    ...         if methodname == "mult":
    ...             return [prefix+"b"]
    ...         else:
    ...             raise KeyError()
    >>> B = A(4)
    >>> m = get_pure_function(B.mult)
    >>> m.set_objparams([3])
    >>> m(2)
    6

    """

    def __init__(self, obj: EditableModule, method: Callable):
        """Initialize the EditableModulePureFunction.

        Parameters
        ----------
        obj: EditableModule
            The object to be wrapped
        method: Callable
            The method to be wrapped

        """
        self.obj = obj
        self.method = method
        super().__init__(method)

    def _get_all_obj_params_init(self) -> List:
        """Get the initial object parameters.

        Returns
        -------
        List
            The initial object parameters

        """
        return list(self.obj.getparams(self.method.__name__))

    def _set_all_obj_params(self, allobjparams: List):
        """Set the object parameters.

        Parameters
        ----------
        allobjparams: List
            The object parameters to be set

        """
        self.obj.setparams(self.method.__name__, *allobjparams)


class TorchNNPureFunction(PureFunction):
    """Implementation of PureFunction for torch.nn.Module.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import get_pure_function
    >>> class A(torch.nn.Module):
    ...     def __init__(self, a):
    ...         super().__init__()
    ...         self.b = torch.nn.Parameter(torch.tensor(a*a))
    ...     def forward(self, x):
    ...         return self.b * x
    >>> B = A(4.)
    >>> m = get_pure_function(B.forward)
    >>> m.set_objparams([3.])
    >>> m(2)
    6.0

    """

    def __init__(self, obj: torch.nn.Module, method: Callable):
        """Initialize the TorchNNPureFunction.

        Parameters
        ----------
        obj: torch.nn.Module
            Object to be wrapped
        method: Callable
            Method to be wrapped

        """
        self.obj = obj
        self.method = method
        super().__init__(method)

    def _get_all_obj_params_init(self) -> List:
        """get the tensors in the torch.nn.Module to be used as params

        Returns
        -------
        List
            The initial object parameters

        """
        named_params = list(self.obj.named_parameters())
        if len(named_params) == 0:
            paramnames: List[str] = []
            obj_params: List[Union[torch.Tensor, torch.nn.Parameter]] = []
        else:
            paramnames_temp, obj_params_temp = zip(*named_params)
            paramnames = list(paramnames_temp)
            obj_params = list(obj_params_temp)
        self.names = paramnames
        return obj_params

    def _set_all_obj_params(self, objparams: List):
        """Set the object parameters.

        Parameters
        ----------
        objparams: List
            The object parameters to be set

        """
        for (name, param) in zip(self.names, objparams):
            del_attr(
                self.obj, name
            )  # delete required in case the param is not a torch.nn.Parameter
            set_attr(self.obj, name, param)


def _check_identical_objs(objs1: List, objs2: List) -> bool:
    """Check if the two lists of objects are identical.

    Examples
    --------
    >>> l1 = [2, 2, 3]
    >>> l2 = [1, 2, 3]
    >>> _check_identical_objs(l1, l2)
    False

    Parameters
    ----------
    objs1: List
        The first list of objects
    objs2: List
        The second list of objects

    Returns
    -------
    bool
        True if the two lists of objects are identical, False otherwise

    """
    for obj1, obj2 in zip(objs1, objs2):
        if id(obj1) != id(obj2):
            return False
    return True


class SingleSiblingPureFunction(PureFunction):
    """Implementation of PureFunction for a sibling method

    A sibling method is a method that is virtually belong to the same object,
    but behaves differently.

    Changing the state of the decorated function will also change the state of
    ``pfunc`` and its other siblings.

    """

    def __init__(self, fcn: Callable, fcntocall: Callable):
        """Initialize the SingleSiblingPureFunction.

        Parameters
        ----------
        fcn: Callable
            The sibling method to be wrapped
        fcntocall: Callable
            The method to be wrapped

        """
        self.pfunc = get_pure_function(fcn)
        super().__init__(fcntocall)

    def _get_all_obj_params_init(self) -> List:
        """Get the initial object parameters.

        Returns
        -------
        List
            The initial object parameters

        """
        return self.pfunc._get_all_obj_params_init()

    def _set_all_obj_params(self, allobjparams: List):
        """Set the object parameters.

        Parameters
        ----------
        allobjparams: List
            The object parameters to be set

        """
        self.pfunc._set_all_obj_params(allobjparams)


class MultiSiblingPureFunction(PureFunction):
    """Implementation of PureFunction for multiple sibling methods

    A sibling method is a method that is virtually belong to the same object,
    but behaves differently.

    Changing the state of the decorated function will also change the state of
    ``pfunc`` and its other siblings.

    """

    def __init__(self, fcns: Sequence[Callable], fcntocall: Callable):
        """Initialize the MultiSiblingPureFunction.

        Parameters
        ----------
        fcns: Sequence[Callable]
            The sibling methods to be wrapped
        fcntocall: Callable
            The method to be wrapped

        """
        self.pfuncs = [get_pure_function(fcn) for fcn in fcns]
        self.npfuncs = len(self.pfuncs)
        super().__init__(fcntocall)

    def _get_all_obj_params_init(self) -> List:
        """Get the initial object parameters.

        Returns
        -------
        List
            The initial object parameters

        """
        res: List[Union[torch.Tensor, torch.nn.Parameter]] = []
        self.cumsum_idx = [0] * (self.npfuncs + 1)
        for i, pfunc in enumerate(self.pfuncs):
            objparams = pfunc._get_all_obj_params_init()
            res = res + objparams
            self.cumsum_idx[i + 1] = self.cumsum_idx[i] + len(objparams)
        return res

    def _set_all_obj_params(self, allobjparams: List):
        """Set the object parameters.

        Parameters
        ----------
        allobjparams: List
            The object parameters to be set

        """
        for i, pfunc in enumerate(self.pfuncs):
            pfunc._set_all_obj_params(
                allobjparams[self.cumsum_idx[i]:self.cumsum_idx[i + 1]])


def get_pure_function(fcn) -> PureFunction:
    """Get the pure function form of the function or method ``fcn``.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import get_pure_function
    >>> def fcn(x, y):
    ...     return x + y
    >>> pfunc = get_pure_function(fcn)
    >>> pfunc(1, 2)
    3

    Parameters
    ----------
    fcn: function or method
        Function or method to be converted into a ``PureFunction`` by exposing
        the hidden parameters affecting its outputs.

    Returns
    -------
    PureFunction
        The pure function wrapper

    """

    errmsg = "The input function must be a function, a method of " \
        "torch.nn.Module, a method of xitorch.EditableModule, or a sibling method"

    if isinstance(fcn, PureFunction):
        return fcn

    elif inspect.isfunction(fcn) or isinstance(fcn, torch.jit.ScriptFunction):
        return FunctionPureFunction(fcn)

    # if it is a method from an object, unroll the parameters and add
    # the object's parameters as well
    elif inspect.ismethod(fcn) or hasattr(fcn, "__call__"):
        if inspect.ismethod(fcn):
            obj = fcn.__self__
        else:
            obj = fcn
            fcn = fcn.__call__

        if isinstance(obj, EditableModule):
            return EditableModulePureFunction(obj, fcn)
        elif isinstance(obj, torch.nn.Module):
            return TorchNNPureFunction(obj, fcn)
        else:
            raise RuntimeError(errmsg)

    else:
        raise RuntimeError(errmsg)


def make_sibling(*pfuncs) -> Callable[[Callable], PureFunction]:
    """
    Used as a decor to mark the decorated function as a sibling method of the
    input ``pfunc``.
    Sibling method is a method that is virtually belong to the same object, but
    behaves differently.
    Changing the state of the decorated function will also change the state of
    ``pfunc`` and its other siblings.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import make_sibling
    >>> def fcn1(x, y):
    ...     return x + y
    >>> def fcn2(x, y):
    ...     return x - y
    >>> pfunc1 = get_pure_function(fcn1)
    >>> pfunc2 = get_pure_function(fcn2)
    >>> @make_sibling(pfunc1)
    ... def fcn3(x, y):
    ...     return x * y
    >>> pfunc3(1, 2)
    2

    Parameters
    ----------
    pfuncs: List[Callable]
        The sibling methods to be wrapped

    Returns
    -------
    Callable[[Callable], PureFunction]
        The decorator function

    """
    if len(pfuncs) == 0:
        raise TypeError("At least 1 function is required as the argument")
    elif len(pfuncs) == 1:
        return lambda fcn: SingleSiblingPureFunction(pfuncs[0], fcntocall=fcn)
    else:
        return lambda fcn: MultiSiblingPureFunction(pfuncs, fcntocall=fcn)



from typing import Union, Tuple
import torch
from deepchem.utils.dft_utils import AtomZsType, AtomPosType
from deepchem.utils import get_atomz

__all__ = ["parse_moldesc"]


def parse_moldesc(
    moldesc: Union[str, Tuple[AtomZsType, AtomPosType]],
    dtype: torch.dtype = torch.float64,
    device: torch.device = torch.device('cpu')
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Parse the string of molecular descriptor and returns tensors of atomzs and
    atom positions.
    Examples
    --------
    >>> from deepchem.utils.dft_utils import parse_moldesc
    >>> system = {
    ...     'type': 'mol',
    ...     'kwargs': {
    ...         'moldesc': 'H 0.86625 0 0; F -0.86625 0 0',
    ...         'basis': '6-311++G(3df,3pd)'
    ...     }
    ... }
    >>> atomzs, atomposs = parse_moldesc(system["kwargs"]["moldesc"])
    >>> atomzs
    tensor([1., 9.], dtype=torch.float64)
    >>> atomposs
    tensor([[ 0.8662,  0.0000,  0.0000],
            [-0.8662,  0.0000,  0.0000]], dtype=torch.float64)

    Parameters
    ----------
    moldesc: Union[str, Tuple[AtomZsType, AtomPosType]]
        String that describes the system, e.g. ``"H -1 0 0; H 1 0 0"`` for H2
        molecule separated by 2 Bohr.
    dtype: torch.dtype (default torch.float64)
        The datatype of the returned atomic positions.
    device: torch.device (default torch.device('cpu'))
        The device to store the returned tensors.
    Returns
    -------
    atomzs: torch.Tensor
        The tensor of atomzs [Atom Number].
    atompos: torch.Tensor
        The tensor of atomic positions [Bohr].
    """
    if isinstance(moldesc, str):
        # TODO: use regex!
        elmts = [[
            get_atomz(c.strip()) if i == 0 else float(c.strip())
            for i, c in enumerate(line.split())
        ]
                 for line in moldesc.split(";")]
        atomzs = torch.tensor([line[0] for line in elmts], device=device)
        atompos = torch.tensor([line[1:] for line in elmts],
                               dtype=dtype,
                               device=device)

    else:  # tuple of atomzs, atomposs
        atomzs_raw, atompos_raw = moldesc
        assert len(atomzs_raw) == len(
            atompos_raw), "Mismatch length of atomz and atompos"
        assert len(atomzs_raw) > 0, "Empty atom list"

        # convert the atomz to tensor
        if not isinstance(atomzs_raw, torch.Tensor):
            atomzs = torch.tensor([get_atomz(at) for at in atomzs_raw],
                                  device=device)
        else:
            atomzs = atomzs_raw.to(device)  # already a tensor

        # convert the atompos to tensor
        if not isinstance(atompos_raw, torch.Tensor):
            atompos = torch.as_tensor(atompos_raw, dtype=dtype, device=device)
        else:
            atompos = atompos_raw.to(dtype).to(device)  # already a tensor

    # convert to dtype if atomzs is a floating point tensor, not an integer tensor
    atomzs = atomzs.to(dtype)

    return atomzs, atompos



from __future__ import annotations
from abc import abstractmethod, abstractproperty
import torch
from typing import List, Union, Optional, Tuple
from deepchem.utils.differentiation_utils import EditableModule
from deepchem.utils.dft_utils import BaseHamilton, BaseGrid, SpinParam, ZType, BasisInpType


class BaseSystem(EditableModule):
    """
    System is a class describing the environment before doing the quantum
    chemistry calculation. It contains the information of the atoms, the
    external electric field, the spin, the charge, etc. It also contains the
    Hamiltonian object and the grid object for the calculation. The system
    object is also responsible for setting up the cache for the parameters
    that can be read/written from/to the cache file.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import BaseSystem
    >>> from deepchem.utils.dft_utils import BaseHamilton
    >>> from deepchem.utils.dft_utils import BaseGrid
    >>> class MySystem(BaseSystem):
    ...     def __init__(self):
    ...         self.hamiltonian = BaseHamilton()
    ...         self.grid = BaseGrid()
    ...     def get_hamiltonian(self):
    ...         return self.hamiltonian
    ...     def get_grid(self):
    ...         return self.grid
    ...     def requires_grid(self):
    ...         return True
    >>> system = MySystem()
    >>> system.requires_grid()
    True

    """

    @abstractmethod
    def densityfit(self,
                   method: Optional[str] = None,
                   auxbasis: Optional[BasisInpType] = None) -> BaseSystem:
        """Indicate that the system's Hamiltonian will use density fitting.

        Parameters
        ----------
        method: Optional[str] (default None)
            The density fitting method to use.
        auxbasis: Optional[BasisInpType] (default None)
            Auxiliary basis set to use for density fitting.

        Returns
        -------
        BaseSystem
            The system with density fitting enabled.

        """
        pass

    @abstractmethod
    def get_hamiltonian(self) -> BaseHamilton:
        """Hamiltonian object for the system.

        Returns
        -------
        BaseHamilton
            Hamiltonian object for the system.

        """
        pass

    @abstractmethod
    def set_cache(self,
                  fname: str,
                  paramnames: Optional[List[str]] = None) -> BaseSystem:
        """
        Set up the cache to read/write some parameters from the given files.
        If paramnames is not given, then read/write all cache-able parameters
        specified by each class.

        Parameters
        ----------
        fname: str
            The file name of the cache file.
        paramnames: Optional[List[str]] (default None)
            The list of parameter names to read/write from the cache file.

        Returns
        -------
        BaseSystem
            The system with cache enabled.

        """
        pass

    @abstractmethod
    def get_orbweight(
        self,
        polarized: bool = False
    ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """
        Returns the atomic orbital weights. If polarized == False, then it
        returns the total orbital weights. Otherwise, it returns a tuple of
        orbital weights for spin-up and spin-down.

        Parameters
        ----------
        polarized: bool (default False)
            Whether to return the orbital weights for each spin.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            The orbital weights. Shape (*BS, norb)

        """
        pass

    @abstractmethod
    def get_nuclei_energy(self) -> torch.Tensor:
        """Returns the nuclei-nuclei repulsion energy.

        Returns
        -------
        torch.Tensor
            The nuclei-nuclei repulsion energy.

        """
        pass

    @abstractmethod
    def setup_grid(self) -> None:
        """Construct the integration grid for the system."""
        pass

    @abstractmethod
    def get_grid(self) -> BaseGrid:
        """
        Returns the grid of the system
        """
        pass

    @abstractmethod
    def requires_grid(self) -> bool:
        """
        True if the system needs the grid to be constructed. Otherwise, returns
        False
        """
        pass

    @abstractmethod
    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Return a list with the parameter names corresponding to the given method
        (methodname)

        Parameters
        ----------
        methodname: str
            The name of the method.
        prefix: str (default "")
            The prefix of the parameter names.

        Returns
        -------
        List[str]
            List of parameter names of methodname

        """
        pass

    @abstractmethod
    def make_copy(self, **kwargs) -> BaseSystem:
        """Copy of the system identical to the orginal except for new parameters
        set in the kwargs.

        Parameters
        ----------
        kwargs
            New parameters to set in the copy.

        Returns
        -------
        BaseSystem
            Copy of the system identical to the orginal except for new parameters
            set in the kwargs.

        """
        pass

    # System Properties
    @abstractproperty
    def atompos(self) -> torch.Tensor:
        """Atom positions with shape ``(natoms, ndim)``.

        Returns
        -------
        torch.Tensor
            Atom positions with shape ``(natoms, ndim)``.

        """
        pass

    @abstractproperty
    def atomzs(self) -> torch.Tensor:
        """Atomic number with shape ``(natoms,)``.

        Returns
        -------
        torch.Tensor
            Atomic number with shape ``(natoms,)``.

        """
        pass

    @abstractproperty
    def atommasses(self) -> torch.Tensor:
        """Atomic mass with shape ``(natoms)`` in atomic unit.

        Returns
        -------
        torch.Tensor
            Atomic mass with shape ``(natoms)`` in atomic unit.

        """
        pass

    @abstractproperty
    def spin(self) -> ZType:
        """Total spin of the system.

        Returns
        -------
        ZType
            Total spin of the system.

        """
        pass

    @abstractproperty
    def charge(self) -> ZType:
        """Charge of the system.

        Returns
        -------
        ZType
            Charge of the system.

        """
        pass

    @abstractproperty
    def numel(self) -> ZType:
        """Total number of the electrons in the system.

        Returns
        -------
        ZType
            Total number of the electrons in the system.

        """
        pass

    @abstractproperty
    def efield(self) -> Optional[Tuple[torch.Tensor, ...]]:
        """
        External electric field of the system, or None if there is no electric
        field.
        """
        pass



from typing import List
try:
    import torch
    import torch.nn as nn
except ModuleNotFoundError:
    raise ImportError('The module requires PyTorch to be installed')

from deepchem.models.torch_models.attention import SelfAttention


class GroverReadout(nn.Module):
    """Performs readout on a batch of graph

    The readout module is used for performing readouts on batched graphs to
    convert node embeddings/edge embeddings into graph embeddings. It is used
    in the Grover architecture to generate a graph embedding from node and edge
    embeddings. The generate embedding can be used in downstream tasks like graph
    classification or graph prediction problems.

    Parameters
    ----------
    rtype: str
        Readout type, can be 'mean' or 'self-attention'
    in_features: int
        Size fof input features
    attn_hidden_size: int
        If readout type is attention, size of hidden layer in attention network.
    attn_out_size: int
        If readout type is attention, size of attention out layer.

    Example
    -------
    >>> import torch
    >>> from deepchem.models.torch_models.readout import GroverReadout
    >>> n_nodes, n_features = 6, 32
    >>> readout = GroverReadout(rtype="mean")
    >>> embedding = torch.ones(n_nodes, n_features)
    >>> result = readout(embedding, scope=[(0, 6)])
    >>> result.size()
    torch.Size([1, 32])
    """

    def __init__(self,
                 rtype: str = 'mean',
                 in_features: int = 128,
                 attn_hidden_size: int = 32,
                 attn_out_size: int = 32):
        super(GroverReadout, self).__init__()
        self.cached_zero_vector = nn.Parameter(torch.zeros(in_features),
                                               requires_grad=False)
        self.rtype = rtype
        if rtype == "self_attention":
            self.attn = SelfAttention(hidden_size=attn_hidden_size,
                                      in_features=in_features,
                                      out_features=attn_out_size)

    def forward(self, graph_embeddings: torch.Tensor,
                scope: List[List]) -> torch.Tensor:
        """Given a batch node/edge embedding and a scope list, produce the graph-level embedding by scope.

        Parameters
        ----------
        embeddings: torch.Tensor
            The embedding matrix, num_nodes x in_features or num_edges x in_features.
        scope: List[List]
            A list, in which the element is a list [start, range]. `start` is the index,
            `range` is the length of scope. (start + range = end)

        Returns
        ----------
        graph_embeddings: torch.Tensor
            A stacked tensor containing graph embeddings of shape len(scope) x in_features if readout type is mean or len(scope) x attn_out_size when readout type is self-attention.
        """
        embeddings: List[torch.Tensor] = []
        for _, (a_start, a_size) in enumerate(scope):
            if a_size == 0:
                embeddings.append(self.cached_zero_vector)
            else:
                embedding = graph_embeddings.narrow(0, a_start, a_size)
                if self.rtype == "self_attention":
                    embedding, attn = self.attn(embedding)
                    embedding = embedding.flatten()
                elif self.rtype == "mean":
                    embedding = embedding.sum(dim=0) / a_size
                embeddings.append(embedding)

        graph_embeddings = torch.stack(embeddings, dim=0)
        return graph_embeddings



from deepchem.feat import Featurizer
from typing import Dict, List
try:
    from transformers import RobertaTokenizerFast
except ModuleNotFoundError:
    raise ImportError(
        'Transformers must be installed for RobertaFeaturizer to be used!')
    pass


class RobertaFeaturizer(RobertaTokenizerFast, Featurizer):
    """Roberta Featurizer.

    The Roberta Featurizer is a wrapper class of the Roberta Tokenizer,
    which is used by Huggingface's transformers library for tokenizing large corpuses for Roberta Models.
    Please confirm the details in [1]_.

    Please see https://github.com/huggingface/transformers
    and https://github.com/seyonechithrananda/bert-loves-chemistry for more details.

    Examples
    --------
    >>> from deepchem.feat import RobertaFeaturizer
    >>> smiles = ["Cn1c(=O)c2c(ncn2C)n(C)c1=O", "CC(=O)N1CN(C(C)=O)C(O)C1O"]
    >>> featurizer = RobertaFeaturizer.from_pretrained("seyonec/SMILES_tokenized_PubChem_shard00_160k")
    >>> out = featurizer(smiles, add_special_tokens=True, truncation=True)

    References
    ----------
    .. [1] Chithrananda, Seyone, Grand, Gabriel, and Ramsundar, Bharath (2020): "Chemberta: Large-scale self-supervised
        pretraining for molecular property prediction." arXiv. preprint. arXiv:2010.09885.


    Note
    -----
    This class requires transformers to be installed.
    RobertaFeaturizer uses dual inheritance with RobertaTokenizerFast in Huggingface for rapid tokenization,
    as well as DeepChem's MolecularFeaturizer class.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        return

    def _featurize(self, datapoint: str, **kwargs) -> List[List[int]]:
        """Calculate encoding using HuggingFace's RobertaTokenizerFast

        Parameters
        ----------
        sequence: str
            Arbitrary string sequence to be tokenized.

        Returns
        -------
        datapoint: List
            List containing two lists; the `input_ids` and the `attention_mask`
        """

        # the encoding is natively a dictionary with keys 'input_ids' and 'attention_mask'
        encoding = list(self(datapoint, **kwargs).values())
        return encoding

    def __call__(self, *args, **kwargs) -> Dict[str, List[int]]:
        return super().__call__(*args, **kwargs)



import numpy as np
import torch
import warnings
from typing import Sequence, Tuple, Union, Optional, Callable, Mapping, Any
from deepchem.utils.differentiation_utils import LinearOperator, MatrixLinearOperator, normalize_bcast_dims, get_bcasted_dims, set_default_option, dummy_context_manager, get_method
from deepchem.utils import ConvergenceWarning, get_np_dtype
from scipy.sparse.linalg import gmres as scipy_gmres


def solve(A: LinearOperator,
          B: torch.Tensor,
          E: Union[torch.Tensor, None] = None,
          M: Optional[LinearOperator] = None,
          bck_options: Mapping[str, Any] = {},
          method: Union[str, Callable, None] = None,
          **fwd_options) -> torch.Tensor:
    r"""
    Performing iterative method to solve the equation.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    >>> B = torch.tensor([[5., 6], [7, 8]])
    >>> solve(A, B)
    tensor([[-3., -4.],
            [ 4.,  5.]])

    .. math::

        \mathbf{AX=B}

    or

    .. math::

        \mathbf{AX-MXE=B}

    where :math:`\mathbf{E}` is a diagonal matrix.
    This function can also solve batched multiple inverse equation at the
    same time by applying :math:`\mathbf{A}` to a tensor :math:`\mathbf{X}`
    with shape ``(...,na,ncols)``.
    The applied :math:`\mathbf{E}` are not necessarily identical for each column.

    Parameters
    ----------
    A: LinearOperator
        A linear operator that takes an input ``X`` and produce the vectors in the same
        space as ``B``.
        It should have the shape of ``(*BA, na, na)``
    B: torch.Tensor
        The tensor on the right hand side with shape ``(*BB, na, ncols)``
    E: Union[torch.Tensor, None]
        If a tensor, it will solve :math:`\mathbf{AX-MXE = B}`.
        It will be regarded as the diagonal of the matrix.
        Otherwise, it just solves :math:`\mathbf{AX = B}` and ``M`` is ignored.
        If it is a tensor, it should have shape of ``(*BE, ncols)``.
    M: Optional[LinearOperator]
        The transformation on the ``E`` side. If ``E`` is ``None``,
        then this argument is ignored.
        If E is not ``None`` and ``M`` is ``None``, then ``M=I``.
        If LinearOperator, it must be Hermitian with shape ``(*BM, na, na)``.
    bck_options: dict
        Options of the iterative solver in the backward calculation.
    method: Union[str, Callable, None]
        The method of linear equation solver. If ``None``, it will choose
        ``"cg"`` or ``"bicgstab"`` based on the matrices symmetry.
        `Note`: default method will be changed quite frequently, so if you want
        future compatibility, please specify a method.
    **fwd_options
        Method-specific options

    Returns
    -------
    torch.Tensor
        The tensor :math:`\mathbf{X}` that satisfies :math:`\mathbf{AX-MXE=B}`.

    """
    assert A.shape[-1] == A.shape[
        -2], "The linear operator A must have a square shape"
    assert A.shape[-1] == B.shape[
        -2], "Mismatch shape of A & B (A: %s, B: %s)" % (A.shape, B.shape)
    assert not torch.is_grad_enabled() or A.is_getparamnames_implemented, \
        "The _getparamnames(self, prefix) of linear operator A must be "\
        "implemented if using solve with grad enabled"
    if M is not None:
        assert M.shape[-1] == M.shape[
            -2], "The linear operator M must have a square shape"
        assert M.shape[-1] == A.shape[
            -1], "The shape of A & M must match (A: %s, M: %s)" % (A.shape,
                                                                   M.shape)
        assert M.is_hermitian, "The linear operator M must be a Hermitian matrix"
        assert not torch.is_grad_enabled() or M.is_getparamnames_implemented, \
            "The _getparamnames(self, prefix) of linear operator M must be "\
            "implemented if using solve with grad enabled"
    if E is not None:
        assert E.shape[-1] == B.shape[-1], \
                        "The last dimension of E & B must match (E: %s, B: %s)" % (E.shape, B.shape)
    if E is None and M is not None:
        warnings.warn(
            "M is supplied but will be ignored because E is not supplied")

    if method is None:
        if isinstance(A, MatrixLinearOperator) and \
           (M is None or isinstance(M, MatrixLinearOperator)):
            method = "exactsolve"
        elif A.shape[-1] <= 5:  # for small matrix
            method = "exactsolve"
        else:
            is_hermit = A.is_hermitian and (M is None or M.is_hermitian)
            method = "cg" if is_hermit else "bicgstab"

    if method == "exactsolve":
        return exactsolve(A, B, E, M)
    else:
        # get the unique parameters of A
        params = A.getlinopparams()
        mparams = M.getlinopparams() if M is not None else []
        na = len(params)
        return solve_torchfcn.apply(A, B, E, M, method, fwd_options,
                                    bck_options, na, *params, *mparams)


class solve_torchfcn(torch.autograd.Function):

    @staticmethod
    def forward(ctx, A, B, E, M, method, fwd_options, bck_options, na,
                *all_params):
        """Forward calculation of the solve function.

        Parameters
        ----------
        A: LinearOperator
            A linear operator that takes an input ``X`` and produce the vectors in the same
            space as ``B``.
            It should have the shape of ``(*BA, na, na)``
        B: torch.Tensor
            The tensor on the right hand side with shape ``(*BB, na, ncols)``
        E: Union[torch.Tensor, None]
            If a tensor, it will solve :math:`\mathbf{AX-MXE = B}`.
            It will be regarded as the diagonal of the matrix.
            Otherwise, it just solves :math:`\mathbf{AX = B}` and ``M`` is ignored.
            If it is a tensor, it should have shape of ``(*BE, ncols)``.
        M: Optional[LinearOperator]
            The transformation on the ``E`` side. If ``E`` is ``None``,
            then this argument is ignored.
            If E is not ``None`` and ``M`` is ``None``, then ``M=I``.
        method: Union[str, Callable, None]
            The method of linear equation solver. If ``None``, it will choose
            ``"cg"`` or ``"bicgstab"`` based on the matrices symmetry.
            `Note`: default method will be changed quite frequently, so if you want
            future compatibility, please specify a method.
        fwd_options
            Method-specific options
        bck_options: dict
            Options of the iterative solver in the backward calculation.
        na: int
            Number of parameters of A
        all_params: Sequence[torch.Tensor]
            All the parameters of M and A

        """

        # separate the parameters for A and for M
        params = all_params[:na]
        mparams = all_params[na:]

        config = set_default_option({}, fwd_options)
        ctx.bck_config = set_default_option({}, bck_options)

        if torch.all(B == 0):  # special case
            dims = (*get_batchdims(A, B, E, M), *B.shape[-2:])
            x = torch.zeros(dims, dtype=B.dtype, device=B.device)
        else:
            with A.uselinopparams(*params), M.uselinopparams(
                    *mparams) if M is not None else dummy_context_manager():
                methods = {
                    "exactsolve": exactsolve,
                    "scipy_gmres": wrap_gmres,
                    "cg": cg,
                    "bicgstab": bicgstab,
                    "gmres": gmres,
                }
                method_fcn = get_method("solve", methods, method)
                x = method_fcn(A, B, E, M, **config)

        ctx.e_is_none = E is None
        ctx.A = A
        ctx.M = M
        if ctx.e_is_none:
            ctx.save_for_backward(x, *all_params)
        else:
            ctx.save_for_backward(x, E, *all_params)
        ctx.na = na
        return x

    @staticmethod
    def backward(ctx, grad_x):
        # grad_x: (*BABEM, nr, ncols)
        # x: (*BABEM, nr, ncols)
        x = ctx.saved_tensors[0]
        idx_all_params = 1 if ctx.e_is_none else 2
        all_params = ctx.saved_tensors[idx_all_params:]
        params = all_params[:ctx.na]
        mparams = all_params[ctx.na:]
        E = None if ctx.e_is_none else ctx.saved_tensors[1]

        # solve (A-biases*M)^T v = grad_x
        # this is the grad of B
        with ctx.A.uselinopparams(*params), \
             ctx.M.uselinopparams(*mparams) if ctx.M is not None else dummy_context_manager():
            AT = ctx.A.H  # (*BA, nr, nr)
            MT = ctx.M.H if ctx.M is not None else None  # (*BM, nr, nr)
            Econj = E.conj() if E is not None else None
            v = solve(AT,
                      grad_x,
                      Econj,
                      MT,
                      bck_options=ctx.bck_config,
                      **ctx.bck_config)  # (*BABEM, nr, ncols)
        grad_B = v

        # calculate the grad of matrices parameters
        with torch.enable_grad():
            params = [p.clone().requires_grad_() for p in params]
            with ctx.A.uselinopparams(*params):
                loss = -ctx.A.mm(x)  # (*BABEM, nr, ncols)

        grad_params = torch.autograd.grad((loss,),
                                          params,
                                          grad_outputs=(v,),
                                          create_graph=torch.is_grad_enabled(),
                                          allow_unused=True)

        # calculate the biases gradient
        grad_E = None
        if E is not None:
            if ctx.M is None:
                Mx = x
            else:
                with ctx.M.uselinopparams(*mparams):
                    Mx = ctx.M.mm(x)  # (*BABEM, nr, ncols)
            grad_E = torch.einsum('...rc,...rc->...c', v,
                                  Mx.conj())  # (*BABEM, ncols)

        # calculate the gradient to the biases matrices
        grad_mparams = []
        if ctx.M is not None and E is not None:
            with torch.enable_grad():
                mparams = [p.clone().requires_grad_() for p in mparams]
                lmbdax = x * E.unsqueeze(-2)
                with ctx.M.uselinopparams(*mparams):
                    mloss = ctx.M.mm(lmbdax)

            grad_mparams = torch.autograd.grad(
                (mloss,),
                mparams,
                grad_outputs=(v,),
                create_graph=torch.is_grad_enabled(),
                allow_unused=True)

        return (None, grad_B, grad_E, None, None, None, None, None,
                *grad_params, *grad_mparams)


# Hidden
def wrap_gmres(A, B, E=None, M=None, min_eps=1e-9, max_niter=None, **unused):
    """
    Using SciPy's gmres method to solve the linear equation.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    >>> B = torch.tensor([[[5., 6], [7, 8]]])
    >>> wrap_gmres(A, B, None, None)
    tensor([[[-3.0000, -4.0000],
             [ 4.0000,  5.0000]]])

    Parameters
    ----------
    A: LinearOperator
        The linear operator A to be solved. Shape: (*BA, na, na)
    B: torch.Tensor
        Batched matrix B. Shape: (*BB, na, ncols)
    E: torch.Tensor or None
        Batched vector E. Shape: (*BE, ncols)
    M: LinearOperator or None
        The linear operator M. Shape: (*BM, na, na)
    min_eps: float
        Relative tolerance for stopping conditions
    max_niter: int or None
        Maximum number of iterations. If ``None``, default to twice of the
        number of columns of ``A``.

    Returns
    -------
    torch.Tensor
        The Solution matrix X. Shape: (*BBE, na, ncols)

    """

    # NOTE: currently only works for batched B (1 batch dim), but unbatched A
    assert len(A.shape) == 2 and len(
        B.shape
    ) == 3, "Currently only works for batched B (1 batch dim), but unbatched A"
    assert not torch.is_complex(B), "complex is not supported in gmres"

    # check the parameters
    msg = "GMRES can only do AX=B"
    assert A.shape[-2] == A.shape[
        -1], "GMRES can only work for square operator for now"
    assert E is None, msg
    assert M is None, msg

    nbatch, na, ncols = B.shape
    if max_niter is None:
        max_niter = 2 * na

    B = B.transpose(-1, -2)  # (nbatch, ncols, na)

    # convert the numpy/scipy
    op = A.scipy_linalg_op()
    B_np = B.detach().cpu().numpy()
    res_np = np.empty(B.shape, dtype=get_np_dtype(B.dtype))
    for i in range(nbatch):
        for j in range(ncols):
            x, info = scipy_gmres(op,
                                  B_np[i, j, :],
                                  tol=min_eps,
                                  atol=1e-12,
                                  maxiter=max_niter)
            if info > 0:
                msg = "The GMRES iteration does not converge to the desired value "\
                      "(%.3e) after %d iterations" % \
                      (min_eps, info)
                warnings.warn(ConvergenceWarning(msg))
            res_np[i, j, :] = x

    res = torch.tensor(res_np, dtype=B.dtype, device=B.device)
    res = res.transpose(-1, -2)  # (nbatch, na, ncols)
    return res


def exactsolve(A: LinearOperator, B: torch.Tensor, E: Union[torch.Tensor, None],
               M: Union[LinearOperator, None]):
    """
    Solve the linear equation by contructing the full matrix of LinearOperators.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    >>> B = torch.tensor([[5., 6], [7, 8]])
    >>> exactsolve(A, B, None, None)
    tensor([[-3., -4.],
            [ 4.,  5.]])

    Parameters
    ----------
    A: LinearOperator
        The linear operator A to be solved. Shape: (*BA, na, na)
    B: torch.Tensor
        Batched matrix B. Shape: (*BB, na, ncols)
    E: torch.Tensor or None
        Batched vector E. Shape: (*BE, ncols)
    M: LinearOperator or None
        The linear operator M. Shape: (*BM, na, na)

    Returns
    -------
    torch.Tensor
        The Solution matrix X. Shape: (*BBE, na, ncols)

    Warnings
    --------
    * As this method construct the linear operators explicitly, it might requires
      a large memory.

    """
    if E is None:
        Amatrix = A.fullmatrix()
        x = torch.linalg.solve(Amatrix, B)
    elif M is None:
        Amatrix = A.fullmatrix()
        x = solve_ABE(Amatrix, B, E)
    else:
        Mmatrix = M.fullmatrix()
        L = torch.linalg.cholesky(Mmatrix)
        Linv = torch.inverse(L)
        LinvT = Linv.transpose(-2, -1).conj()
        A2 = torch.matmul(Linv, A.mm(LinvT))
        B2 = torch.matmul(Linv, B)

        X2 = solve_ABE(A2, B2, E)
        x = torch.matmul(LinvT, X2)
    return x


def solve_ABE(A: torch.Tensor, B: torch.Tensor, E: torch.Tensor):
    """ Solve the linear equation AX = B - diag(E)X.

    Examples
    --------
    >>> import torch
    >>> A = torch.tensor([[1., 2], [3, 4]])
    >>> B = torch.tensor([[5., 6], [7, 8]])
    >>> E = torch.tensor([1., 2])
    >>> solve_ABE(A, B, E)
    tensor([[-0.1667,  0.5000],
            [ 2.5000,  3.2500]])

    Parameters
    ----------
    A: torch.Tensor
        The batched matrix A. Shape: (*BA, na, na)
    B: torch.Tensor
        The batched matrix B. Shape: (*BB, na, ncols)
    E: torch.Tensor
        The batched vector E. Shape: (*BE, ncols)

    Returns
    -------
    torch.Tensor
        The batched matrix X.

    """
    na = A.shape[-1]
    BA, BB, BE = normalize_bcast_dims(A.shape[:-2], B.shape[:-2], E.shape[:-1])
    E = E.reshape(1, *BE, E.shape[-1]).transpose(0, -1)  # (ncols, *BE, 1)
    B = B.reshape(1, *BB, *B.shape[-2:]).transpose(0, -1)  # (ncols, *BB, na, 1)

    # NOTE: The line below is very inefficient for large na and ncols
    AE = A - torch.diag_embed(E.repeat_interleave(repeats=na, dim=-1),
                              dim1=-2,
                              dim2=-1)  # (ncols, *BAE, na, na)
    r = torch.linalg.solve(AE, B)  # (ncols, *BAEM, na, 1)
    r = r.transpose(0, -1).squeeze(0)  # (*BAEM, na, ncols)
    return r


def cg(A: LinearOperator,
       B: torch.Tensor,
       E: Optional[torch.Tensor] = None,
       M: Optional[LinearOperator] = None,
       posdef: Optional[bool] = None,
       precond: Optional[LinearOperator] = None,
       max_niter: Optional[int] = None,
       rtol: float = 1e-6,
       atol: float = 1e-8,
       eps: float = 1e-12,
       resid_calc_every: int = 10,
       verbose: bool = False,
       **unused) -> torch.Tensor:
    r"""
    Solve the linear equations using Conjugate-Gradient (CG) method.

    Parameters
    ----------
    A: LinearOperator
        A linear operator that takes an input ``X`` and produce the vectors in the same
        space as ``B``.
        It should have the shape of ``(*BA, na, na)``
    B: torch.Tensor
        The tensor on the right hand side with shape ``(*BB, na, ncols)``
    E: Union[torch.Tensor, None]
        If a tensor, it will solve :math:`\mathbf{AX-MXE = B}`.
        It will be regarded as the diagonal of the matrix.
        Otherwise, it just solves :math:`\mathbf{AX = B}` and ``M`` is ignored.
        If it is a tensor, it should have shape of ``(*BE, ncols)``.
    M: Optional[LinearOperator]
        The transformation on the ``E`` side. If ``E`` is ``None``,
        then this argument is ignored.
        If E is not ``None`` and ``M`` is ``None``, then ``M=I``.
    posdef: bool or None
        Indicating if the operation :math:`\mathbf{AX-MXE}` a positive
        definite for all columns and batches.
        If None, it will be determined by power iterations.
    precond: LinearOperator or None
        LinearOperator for the preconditioning. If None, no preconditioner is
        applied.
    max_niter: int or None
        Maximum number of iteration. If None, it is set to ``int(1.5 * A.shape[-1])``
    rtol: float
        Relative tolerance for stopping condition w.r.t. norm of B
    atol: float
        Absolute tolerance for stopping condition w.r.t. norm of B
    eps: float
        Substitute the absolute zero in the algorithm's denominator with this
        value to avoid nan.
    resid_calc_every: int
        Calculate the residual in its actual form instead of substitution form
        with this frequency, to avoid rounding error accummulation.
        If your linear operator has bad numerical precision, set this to be low.
        If 0, then never calculate the residual in its actual form.
    verbose: bool
        Verbosity of the algorithm.
    """
    nr = A.shape[-1]
    ncols = B.shape[-1]
    if max_niter is None:
        max_niter = int(1.5 * nr)

    # if B is all zeros, then return zeros
    batchdims = get_batchdims(A, B, E, M)
    if torch.allclose(B, B * 0, rtol=rtol, atol=atol):
        x0 = torch.zeros((*batchdims, nr, ncols),
                         dtype=A.dtype,
                         device=A.device)
        return x0

    # setup the preconditioning and the matrix problem
    precond_fcn = setup_precond(precond)
    need_hermit = True
    A_fcn, _, B2, col_swapped = setup_linear_problem(A, B, E, M, batchdims,
                                                     posdef, need_hermit)

    # get the stopping matrix
    B_norm = B2.norm(dim=-2, keepdim=True)  # (*BB, 1, nc)
    stop_matrix = torch.max(rtol * B_norm,
                            atol * torch.ones_like(B_norm))  # (*BB, 1, nc)

    # prepare the initial guess (it's just all zeros)
    x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr,
                                                              ncols)
    xk = torch.zeros(x0shape, dtype=A.dtype, device=A.device)

    rk = B2 - A_fcn(xk)  # (*, nr, nc)
    zk = precond_fcn(rk)  # (*, nr, nc)
    pk = zk  # (*, nr, nc)
    rkzk = dot(rk, zk)
    converge = False
    best_resid = rk.norm(dim=-2).max().item()
    best_xk = xk
    for k in range(1, max_niter + 1):
        Apk = A_fcn(pk)
        alphak = rkzk / safedenom(dot(pk, Apk), eps)
        xk_1 = xk + alphak * pk

        # correct the residual calculation
        if resid_calc_every != 0 and k % resid_calc_every == 0:
            rk_1 = B2 - A_fcn(xk_1)
        else:
            rk_1 = rk - alphak * Apk  # (*, nr, nc)

        # check for the stopping condition
        resid = rk_1  # B2 - A_fcn(xk_1)
        resid_norm = resid.norm(dim=-2, keepdim=True)

        max_resid_norm = resid_norm.max().item()
        if max_resid_norm < best_resid:
            best_resid = max_resid_norm
            best_xk = xk_1

        if verbose:
            if k < 10 or k % 10 == 0:
                print("%4d: |dy|=%.3e" % (k, resid_norm))

        if torch.all(resid_norm < stop_matrix):
            converge = True
            break

        zk_1 = precond_fcn(rk_1)
        rkzk_1 = dot(rk_1, zk_1)
        betak = rkzk_1 / safedenom(rkzk, eps)
        pk_1 = zk_1 + betak * pk

        # move to the next index
        pk = pk_1
        zk = zk_1
        xk = xk_1
        rk = rk_1
        rkzk = rkzk_1

    xk_1 = best_xk
    if not converge:
        msg = ("Convergence is not achieved after %d iterations. "
               "Max norm of best resid: %.3e") % (max_niter, best_resid)
        warnings.warn(ConvergenceWarning(msg))
    if col_swapped:
        # x: (ncols, *, nr, 1)
        xk_1 = xk_1.transpose(0, -1).squeeze(0)  # (*, nr, ncols)
    return xk_1


def bicgstab(A: LinearOperator,
             B: torch.Tensor,
             E: Optional[torch.Tensor] = None,
             M: Optional[LinearOperator] = None,
             posdef: Optional[bool] = None,
             precond_l: Optional[LinearOperator] = None,
             precond_r: Optional[LinearOperator] = None,
             max_niter: Optional[int] = None,
             rtol: float = 1e-6,
             atol: float = 1e-8,
             eps: float = 1e-12,
             verbose: bool = False,
             resid_calc_every: int = 10,
             **unused) -> torch.Tensor:
    r"""
    Solve the linear equations using stabilized Biconjugate-Gradient method.

    Parameters
    ----------
    posdef: bool or None
        Indicating if the operation :math:`\mathbf{AX-MXE}` a positive
        definite for all columns and batches.
        If None, it will be determined by power iterations.
    precond_l: LinearOperator or None
        LinearOperator for the left preconditioning. If None, no
        preconditioner is applied.
    precond_r: LinearOperator or None
        LinearOperator for the right preconditioning. If None, no
        preconditioner is applied.
    max_niter: int or None
        Maximum number of iteration. If None, it is set to ``int(1.5 * A.shape[-1])``
    rtol: float
        Relative tolerance for stopping condition w.r.t. norm of B
    atol: float
        Absolute tolerance for stopping condition w.r.t. norm of B
    eps: float
        Substitute the absolute zero in the algorithm's denominator with this
        value to avoid nan.
    resid_calc_every: int
        Calculate the residual in its actual form instead of substitution form
        with this frequency, to avoid rounding error accummulation.
        If your linear operator has bad numerical precision, set this to be low.
        If 0, then never calculate the residual in its actual form.
    verbose: bool
        Verbosity of the algorithm.

    """
    nr, ncols = B.shape[-2:]
    if max_niter is None:
        max_niter = int(1.5 * nr)

    # if B is all zeros, then return zeros
    batchdims = get_batchdims(A, B, E, M)
    if torch.allclose(B, B * 0, rtol=rtol, atol=atol):
        x0 = torch.zeros((*batchdims, nr, ncols),
                         dtype=A.dtype,
                         device=A.device)
        return x0

    # setup the preconditioning and the matrix problem
    precond_fcn_l = setup_precond(precond_l)
    precond_fcn_r = setup_precond(precond_r)
    need_hermit = False
    A_fcn, AT_fcn, B2, col_swapped = setup_linear_problem(
        A, B, E, M, batchdims, posdef, need_hermit)

    # get the stopping matrix
    B_norm = B2.norm(dim=-2, keepdim=True)  # (*BB, 1, nc)
    stop_matrix = torch.max(rtol * B_norm,
                            atol * torch.ones_like(B_norm))  # (*BB, 1, nc)

    # prepare the initial guess (it's just all zeros)
    x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr,
                                                              ncols)
    xk = torch.zeros(x0shape, dtype=A.dtype, device=A.device)

    rk = B2 - A_fcn(xk)
    r0hat = rk
    rho_k = dot(r0hat, rk)
    omega_k = torch.tensor(1.0, dtype=A.dtype, device=A.device)
    alpha: Union[float, torch.Tensor] = 1.0
    vk: Union[float, torch.Tensor] = 0.0
    pk: Union[float, torch.Tensor] = 0.0
    converge = False
    best_resid = rk.norm(dim=-2).max()
    best_xk = xk
    for k in range(1, max_niter + 1):
        rho_knew = dot(r0hat, rk)
        omega_denom = safedenom(omega_k, eps)
        beta = rho_knew / safedenom(rho_k, eps) * (alpha / omega_denom)
        pk = rk + beta * (pk - omega_k * vk)
        y = precond_fcn_r(pk)
        vk = A_fcn(y)
        alpha = rho_knew / safedenom(dot(r0hat, vk), eps)
        h = xk + alpha * y

        s = rk - alpha * vk
        z = precond_fcn_r(s)
        t = A_fcn(z)
        Kt = precond_fcn_l(t)
        omega_k = dot(Kt, precond_fcn_l(s)) / safedenom(dot(Kt, Kt), eps)
        xk = h + omega_k * z

        # correct the residual calculation regularly
        if resid_calc_every != 0 and k % resid_calc_every == 0:
            rk = B2 - A_fcn(xk)
        else:
            rk = s - omega_k * t

        # calculate the residual
        resid = rk
        resid_norm = resid.norm(dim=-2, keepdim=True)

        # save the best results
        max_resid_norm = resid_norm.max().item()
        if max_resid_norm < best_resid:
            best_resid = max_resid_norm
            best_xk = xk

        if verbose:
            if k < 10 or k % 10 == 0:
                print("%4d: |dy|=%.3e" % (k, resid_norm))

        # check for the stopping conditions
        if torch.all(resid_norm < stop_matrix):
            converge = True
            break

        rho_k = rho_knew

    xk = best_xk
    if not converge:
        msg = ("Convergence is not achieved after %d iterations. "
               "Max norm of resid: %.3e") % (max_niter, best_resid)
        warnings.warn(ConvergenceWarning(msg))
    if col_swapped:
        # x: (ncols, *, nr, 1)
        xk = xk.transpose(0, -1).squeeze(0)  # (*, nr, ncols)
    return xk


def gmres(A: LinearOperator,
          B: torch.Tensor,
          E: Optional[torch.Tensor] = None,
          M: Optional[LinearOperator] = None,
          posdef: Optional[bool] = None,
          max_niter: Optional[int] = None,
          rtol: float = 1e-6,
          atol: float = 1e-8,
          eps: float = 1e-12,
          **unused) -> torch.Tensor:
    r"""
    Solve the linear equations using Generalised minial residual method.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.differentiation_utils import LinearOperator
    >>> A = LinearOperator.m(torch.tensor([[1., 2], [3, 4]]))
    >>> B = torch.tensor([[5., 6], [7, 8]])
    >>> gmres(A, B)
    tensor([[0.8959, 1.0697],
            [1.2543, 1.4263]])

    Parameters
    ----------
    A: LinearOperator
        The linear operator A to be solved. Shape: (*BA, na, na)
    B: torch.Tensor
        Batched matrix B. Shape: (*BB, na, ncols)
    E: torch.Tensor or None
        Batched vector E. Shape: (*BE, ncols)
    M: LinearOperator or None
        The linear operator M. Shape: (*BM, na, na)
    posdef: bool or None
        Indicating if the operation :math:`\mathbf{AX-MXE}` a positive
        definite for all columns and batches.
        If None, it will be determined by power iterations.
    max_niter: int or None
        Maximum number of iteration. If None, it is set to ``int(1.5 * A.shape[-1])``
    rtol: float
        Relative tolerance for stopping condition w.r.t. norm of B
    atol: float
        Absolute tolerance for stopping condition w.r.t. norm of B
    eps: float
        Substitute the absolute zero in the algorithm's denominator with this
        value to avoid nan.

    Returns
    -------
    torch.Tensor
        The solution matrix X. Shape: (*BBE, na, ncols)

    """
    converge = False

    nr, ncols = A.shape[-1], B.shape[-1]
    if max_niter is None:
        max_niter = int(nr)

    # if B is all zeros, then return zeros
    batchdims = get_batchdims(A, B, E, M)
    if torch.allclose(B, B * 0, rtol=rtol, atol=atol):
        x0 = torch.zeros((*batchdims, nr, ncols),
                         dtype=A.dtype,
                         device=A.device)
        return x0

    # setup the preconditioning and the matrix problem
    need_hermit = False
    A_fcn, AT_fcn, B2, col_swapped = setup_linear_problem(
        A, B, E, M, batchdims, posdef, need_hermit)

    # get the stopping matrix
    B_norm = B2.norm(dim=-2, keepdim=True)  # (*BB, 1, nc)
    stop_matrix = torch.max(rtol * B_norm,
                            atol * torch.ones_like(B_norm))  # (*BB, 1, nc)

    # prepare the initial guess (it's just all zeros)
    x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr,
                                                              ncols)
    x0 = torch.zeros(x0shape, dtype=A.dtype, device=A.device)

    r = B2 - A_fcn(x0)  # torch.Size([*batch_dims, nr, ncols])
    best_resid = r.norm(dim=-2, keepdim=True)  # / B_norm

    best_resid = best_resid.max().item()
    best_res = x0
    q = torch.empty([max_niter] + list(r.shape), dtype=A.dtype, device=A.device)
    q[0] = r / safedenom(r.norm(dim=-2, keepdim=True),
                         eps)  # torch.Size([*batch_dims, nr, ncols])
    h = torch.zeros((*batchdims, ncols, max_niter + 1, max_niter),
                    dtype=A.dtype,
                    device=A.device)
    h = h.reshape((-1, ncols, max_niter + 1, max_niter))

    for k in range(min(nr, max_niter)):
        y = A_fcn(q[k])  # torch.Size([*batch_dims, nr, ncols])
        for j in range(k + 1):
            h[..., j, k] = dot(q[j], y).reshape(-1, ncols)
            y = y - h[..., j, k].reshape(*batchdims, 1, ncols) * q[j]

        h[..., k + 1, k] = torch.linalg.norm(y, dim=-2)
        if torch.any(h[..., k + 1, k]) != 0 and k != max_niter - 1:
            q[k + 1] = y.reshape(-1, nr, ncols) / h[..., k + 1, k].reshape(
                -1, 1, ncols)
            q[k + 1] = q[k + 1].reshape(*batchdims, nr, ncols)

        b = torch.zeros((*batchdims, ncols, k + 1),
                        dtype=A.dtype,
                        device=A.device)
        b = b.reshape(-1, ncols, k + 1)
        b[..., 0] = torch.linalg.norm(r, dim=-2)
        rk = torch.linalg.lstsq(h[..., :k + 1, :k], b)[0]

        res = torch.empty([])
        for i in range(k):
            res = res + q[i] * rk[..., i].reshape(*batchdims, 1, ncols) + x0 if res.size() \
                else q[i] * rk[..., i].reshape(*batchdims, 1, ncols) + x0
            # res = res * B_norm

        if res.size():
            resid = B2 - A_fcn(res)
            resid_norm = resid.norm(dim=-2, keepdim=True)

            # save the best results
            max_resid_norm = resid_norm.max().item()
            if max_resid_norm < best_resid:
                best_resid = max_resid_norm
                best_res = res

            if torch.all(resid_norm < stop_matrix):
                converge = True
                break

    if not converge:
        msg = ("Convergence is not achieved after %d iterations. "
               "Max norm of resid: %.3e") % (max_niter, best_resid)
        warnings.warn(ConvergenceWarning(msg))

    res = best_res
    return res


# general helpers
def get_batchdims(A: LinearOperator, B: torch.Tensor,
                  E: Union[torch.Tensor, None], M: Union[LinearOperator, None]):
    """Get the batch dimensions of the linear operator and the matrix B

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils import MatrixLinearOperator
    >>> import torch
    >>> A = MatrixLinearOperator(torch.randn(4, 3, 3), True)
    >>> B = torch.randn(3, 3, 2)
    >>> get_batchdims(A, B, None, None)
    [4]

    Parameters
    ----------
    A: LinearOperator
        The linear operator. It can be a batched linear operator.
    B: torch.Tensor
        The matrix B. It can be a batched matrix.
    E: Union[torch.Tensor, None]
        The matrix E. It can be a batched matrix.
    M: Union[LinearOperator, None]
        The linear operator M. It can be a batched linear operator.

    Returns
    -------
    List[int]
        The batch dimensions of the linear operator and the matrix B

    """

    batchdims = [A.shape[:-2], B.shape[:-2]]
    if E is not None:
        batchdims.append(E.shape[:-1])
        if M is not None:
            batchdims.append(M.shape[:-2])
    return get_bcasted_dims(*batchdims)


def setup_precond(
    precond: Optional[LinearOperator] = None
) -> Callable[[torch.Tensor], torch.Tensor]:
    """Setup the preconditioning function

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils import MatrixLinearOperator
    >>> import torch
    >>> A = MatrixLinearOperator(torch.randn(4, 3, 3), True)
    >>> B = torch.randn(4, 3, 2)
    >>> cond = setup_precond(A)
    >>> cond(B).shape
    torch.Size([4, 3, 2])

    Parameters
    ----------
    precond: Optional[LinearOperator]
        The preconditioning linear operator. If None, no preconditioning is
        applied.

    Returns
    -------
    Callable[[torch.Tensor], torch.Tensor]
        The preconditioning function. It takes a tensor and returns a tensor.

    """
    if isinstance(precond, LinearOperator):

        def precond_fcn(x):
            return precond.mm(x)
    elif precond is None:

        def precond_fcn(x):
            return x
    else:
        raise TypeError("precond can only be LinearOperator or None")
    return precond_fcn


def setup_linear_problem(A: LinearOperator, B: torch.Tensor,
                          E: Optional[torch.Tensor], M: Optional[LinearOperator],
                          batchdims: Sequence[int],
                          posdef: Optional[bool],
                          need_hermit: bool) -> \
        Tuple[Callable[[torch.Tensor], torch.Tensor],
              Callable[[torch.Tensor], torch.Tensor],
              torch.Tensor, bool]:
    """Setup the linear problem for solving AX = B

    Examples
    --------
    >>> from deepchem.utils.differentiation_utils import MatrixLinearOperator
    >>> import torch
    >>> A = MatrixLinearOperator(torch.randn(4, 3, 3), True)
    >>> B = torch.randn(4, 3, 2)
    >>> A_fcn, AT_fcn, B_new, col_swapped = setup_linear_problem(A, B, None, None, [4], None, False)
    >>> A_fcn(B).shape
    torch.Size([4, 3, 2])

    Parameters
    ----------
    A: LinearOperator
        The linear operator A. It can be a batched linear operator.
    B: torch.Tensor
        The matrix B. It can be a batched matrix.
    E: Optional[torch.Tensor]
        The matrix E. It can be a batched matrix.
    M: Optional[LinearOperator]
        The linear operator M. It can be a batched linear operator.
    batchdims: Sequence[int]
        The batch dimensions of the linear operator and the matrix B
    posdef: Optional[bool]
        Whether the linear operator is positive definite. If None, it will be
        estimated.
    need_hermit: bool
        Whether the linear operator is Hermitian. If True, it will be estimated.

    Returns
    -------
    Tuple[Callable[[torch.Tensor], torch.Tensor],
          Callable[[torch.Tensor], torch.Tensor],
          torch.Tensor, bool]
        The function A, its transposed function, the matrix B, and whether the
        columns of B are swapped.

    """

    # get the linear operator (including the MXE part)
    if E is None:

        def A_fcn(x):
            return A.mm(x)

        def AT_fcn(x):
            return A.rmm(x)

        B_new = B
        col_swapped = False
    else:
        # A: (*BA, nr, nr) linop
        # B: (*BB, nr, ncols)
        # E: (*BE, ncols)
        # M: (*BM, nr, nr) linop
        if M is None:
            BAs, BBs, BEs = normalize_bcast_dims(A.shape[:-2], B.shape[:-2],
                                                 E.shape[:-1])
        else:
            BAs, BBs, BEs, BMs = normalize_bcast_dims(A.shape[:-2],
                                                      B.shape[:-2],
                                                      E.shape[:-1],
                                                      M.shape[:-2])
        E = E.reshape(*BEs, *E.shape[-1:])
        E_new = E.unsqueeze(0).transpose(-1,
                                         0).unsqueeze(-1)  # (ncols, *BEs, 1, 1)
        B = B.reshape(*BBs, *B.shape[-2:])  # (*BBs, nr, ncols)
        B_new = B.unsqueeze(0).transpose(-1, 0)  # (ncols, *BBs, nr, 1)

        def A_fcn(x):
            # x: (ncols, *BX, nr, 1)
            Ax = A.mm(x)  # (ncols, *BAX, nr, 1)
            Mx = M.mm(x) if M is not None else x  # (ncols, *BMX, nr, 1)
            MxE = Mx * E_new  # (ncols, *BMXE, nr, 1)
            return Ax - MxE

        def AT_fcn(x):
            # x: (ncols, *BX, nr, 1)
            ATx = A.rmm(x)
            MTx = M.rmm(x) if M is not None else x
            MTxE = MTx * E_new
            return ATx - MTxE

        col_swapped = True

    # estimate if it's posdef with power iteration
    if need_hermit:
        is_hermit = A.is_hermitian and (M is None or M.is_hermitian)
        if not is_hermit:
            # set posdef to False to make the operator becomes AT * A so it is
            # hermitian
            posdef = False

    # TODO: the posdef check by largest eival only works for Hermitian/symmetric
    # matrix, but it doesn't always work for non-symmetric matrix.
    # In non-symmetric case, one need to do Cholesky LDL decomposition
    if posdef is None:
        nr, ncols = B.shape[-2:]
        x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims,
                                                                  nr, ncols)
        x0 = torch.randn(x0shape, dtype=A.dtype, device=A.device)
        x0 = x0 / x0.norm(dim=-2, keepdim=True)
        largest_eival = get_largest_eival(A_fcn, x0)  # (*, 1, nc)
        negeival = largest_eival <= 0

        # if the largest eigenvalue is negative, then it's not posdef
        if torch.all(negeival):
            posdef = False

        # otherwise, calculate the lowest eigenvalue to check if it's positive
        else:
            offset = torch.clamp(largest_eival, min=0.0)

            def A_fcn2(x):
                return A_fcn(x) - offset * x

            mostneg_eival = get_largest_eival(A_fcn2, x0)  # (*, 1, nc)
            posdef = bool(
                torch.all(torch.logical_or(-mostneg_eival <= offset,
                                           negeival)).item())

    # get the linear operation if it is not a posdef (A -> AT.A)
    if posdef:
        return A_fcn, AT_fcn, B_new, col_swapped
    else:

        def A_new_fcn(x):
            return AT_fcn(A_fcn(x))

        B2 = AT_fcn(B_new)
        return A_new_fcn, A_new_fcn, B2, col_swapped


# cg and bicgstab helpers
def safedenom(r: torch.Tensor, eps: float) -> torch.Tensor:
    """Make sure the denominator is not zero

    Examples
    --------
    >>> import torch
    >>> r = torch.tensor([[0., 2], [3, 4]])
    >>> safedenom(r, 1e-9)
    tensor([[1.0000e-09, 2.0000e+00],
            [3.0000e+00, 4.0000e+00]])

    Parameters
    ----------
    r: torch.Tensor
        The input tensor. Shape: (*BR, nr, nc)
    eps: float
        The small number to replace the zero denominator

    Returns
    -------
    torch.Tensor
        The tensor with non-zero denominator. Shape: (*BR, nr, nc)

    """
    r[r == 0] = eps
    return r


def dot(r: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    """Dot product of two vectors. r and z must have the same shape.
    Then sums it up across the last dimension.

    Examples
    --------
    >>> import torch
    >>> r = torch.tensor([[1, 2], [3, 4]])
    >>> z = torch.tensor([[5, 6], [7, 8]])
    >>> dot(r, z)
    tensor([[26, 44]])

    Parameters
    ----------
    r: torch.Tensor
        The first vector. Shape: (*BR, nr, nc)
    z: torch.Tensor
        The second vector. Shape: (*BR, nr, nc)

    Returns
    -------
    torch.Tensor
        The dot product of r and z. Shape: (*BR, 1, nc)

    """
    return torch.einsum("...rc,...rc->...c", r.conj(), z).unsqueeze(-2)


def get_largest_eival(Afcn: Callable, x: torch.Tensor) -> torch.Tensor:
    """Get the largest eigenvalue of the linear operator Afcn

    Examples
    --------
    >>> import torch
    >>> def Afcn(x):
    ...     return 10 * x
    >>> x = torch.tensor([[1., 2], [3, 4]])
    >>> get_largest_eival(Afcn, x)
    tensor([[10., 10.]])

    Parameters
    ----------
    Afcn: Callable
        The linear operator A. It takes a tensor and returns a tensor.
    x: torch.Tensor
        The input tensor. Shape: (*, nr, nc)

    Returns
    -------
    torch.Tensor
        The largest eigenvalue. Shape: (*, 1, nc)

    """
    niter = 10
    rtol = 1e-3
    atol = 1e-6
    xnorm_prev = None
    for i in range(niter):
        x = Afcn(x)  # (*, nr, nc)
        xnorm = x.norm(dim=-2, keepdim=True)  # (*, 1, nc)

        # check if xnorm is converging
        if i > 0:
            dnorm = torch.abs(xnorm_prev - xnorm)
            if torch.all(dnorm <= rtol * xnorm + atol):
                break

        xnorm_prev = xnorm
        if i < niter - 1:
            x = x / xnorm
    return xnorm



from typing import Tuple
import torch
import numpy as np
from scipy.special import erfcinv


class Lattice(object):
    """
    Lattice is an object that describe the periodicity of the lattice.
    Note that this object does not know about atoms.
    For the integrated object between the lattice and atoms, please see Sol

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import Lattice
    >>> a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
    >>> lattice = Lattice(a)
    >>> lattice.lattice_vectors()
    tensor([[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]])
    >>> lattice.recip_vectors()
    tensor([[6.2832, 0.0000, 0.0000],
            [0.0000, 6.2832, 0.0000],
            [0.0000, 0.0000, 6.2832]])
    >>> lattice.volume() # volume of the unit cell
    tensor(1.)
    >>> lattice.get_lattice_ls(1.0) # get the neighboring lattice vectors
    tensor([[ 0.,  0., -1.],
            [ 0., -1.,  0.],
            [-1.,  0.,  0.],
            [ 0.,  0.,  0.],
            [ 1.,  0.,  0.],
            [ 0.,  1.,  0.],
            [ 0.,  0.,  1.]])
    >>> lattice.get_gvgrids(6.0) # get the neighboring G-vectors
    (tensor([[ 0.0000,  0.0000, -6.2832],
            [ 0.0000, -6.2832,  0.0000],
            [-6.2832,  0.0000,  0.0000],
            [ 0.0000,  0.0000,  0.0000],
            [ 6.2832,  0.0000,  0.0000],
            [ 0.0000,  6.2832,  0.0000],
            [ 0.0000,  0.0000,  6.2832]]), tensor([1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000]))
    >>> lattice.estimate_ewald_eta(1e-5) # estimate the ewald's sum eta
    1.8

    """

    def __init__(self, a: torch.Tensor):
        """Initialize the lattice object.

        2D or 1D repetition are not implemented yet

        Parameters
        ----------
        a: torch.Tensor
            The lattice vectors with shape (ndim, ndim) with ndim == 3

        """
        assert a.ndim == 2
        assert a.shape[0] == 3
        assert a.shape[-1] == 3
        self.ndim = a.shape[0]
        self.a = a
        self.device = a.device
        self.dtype = a.dtype

    def lattice_vectors(self) -> torch.Tensor:
        """Returns the 3D lattice vectors (nv, ndim) with nv == 3"""
        return self.a

    def recip_vectors(self) -> torch.Tensor:
        """
        Returns the 3D reciprocal vectors with norm == 2 * pi with shape (nv, ndim)
        with nv == 3

        Note: ``torch.det(self.a)`` should not be equal to zero.
        """
        return torch.inverse(self.a.transpose(-2, -1)) * (2 * np.pi)

    def volume(self) -> torch.Tensor:
        """Returns the volume of a lattice."""
        return torch.det(self.a)

    @property
    def params(self) -> Tuple[torch.Tensor, ...]:
        """Returns the list of parameters of this object"""
        return (self.a,)

    def get_lattice_ls(self,
                       rcut: float,
                       exclude_zeros: bool = False) -> torch.Tensor:
        """
        Returns a tensor that contains the coordinates of the neighboring
        lattices.

        Parameters
        ----------
        rcut: float
            The threshold of the distance from the main cell to be included
            in the neighbor.
        exclude_zeros: bool (default: False)
            If True, then it will exclude the vector that are all zeros.

        Returns
        -------
        ls: torch.Tensor
            Tensor with size `(nb, ndim)` containing the coordinates of the
            neighboring cells.

        """
        a = self.lattice_vectors()
        return self._generate_lattice_vectors(a,
                                              rcut,
                                              exclude_zeros=exclude_zeros)

    def get_gvgrids(
            self,
            gcut: float,
            exclude_zeros: bool = False) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Returns a tensor that contains the coordinate in reciprocal space of the
        neighboring Brillouin zones.

        Parameters
        ----------
        gcut: float
            Cut off for generating the G-points.
        exclude_zeros: bool (default: False)
            If True, then it will exclude the vector that are all zeros.

        Returns
        -------
        gvgrids: torch.Tensor
            Tensor with size `(ng, ndim)` containing the G-coordinates of the
            Brillouin zones.
        weights: torch.Tensor
            Tensor with size `(ng)` representing the weights of the G-points.

        """
        b = self.recip_vectors()
        gvgrids = self._generate_lattice_vectors(b,
                                                 gcut,
                                                 exclude_zeros=exclude_zeros)

        # 1 / cell.vol == det(b) / (2 pi)^3
        weights = torch.zeros(gvgrids.shape[0],
                              dtype=self.dtype,
                              device=self.device)
        weights = weights + torch.abs(torch.det(b)) / (2 * np.pi)**3
        return gvgrids, weights

    def estimate_ewald_eta(self, precision: float) -> float:
        """estimate the ewald's sum eta for nuclei interaction energy the
        precision is assumed to be relative precision this formula is obtained
        by estimating the sum as an integral.

        Parameters
        ----------
        precision: float
            The precision of the ewald's sum.

        Returns
        -------
        eta: float
            The estimated eta.

        """
        vol = float(self.volume().detach())
        eta0 = np.sqrt(np.pi) / vol**(1. / 3)
        eta = eta0 * erfcinv(0.5 * precision) / erfcinv(precision)
        return round(eta * 10) / 10  # round to 1 d.p.

    def _generate_lattice_vectors(self, a: torch.Tensor, rcut: float,
                                  exclude_zeros: bool) -> torch.Tensor:
        """generate the lattice vectors of multiply of a within the radius rcut.

        Parameters
        ----------
        a: torch.Tensor
            The lattice vectors with shape (ndim, ndim) with ndim == 3
        rcut: float
            The threshold of the distance from the main cell to be included
            in the neighbor.
        exclude_zeros: bool (default: False)
            If True, then it will exclude the vector that are all zeros.

        Returns
        -------
        ls: torch.Tensor
            Tensor with size `(nb, ndim)` containing the coordinates of the
            neighboring cells.

        Derived From pyscf:
        https://github.com/pyscf/pyscf/blob/e6c569932d5bab5e49994ae3dd365998fc5202b5/pyscf/pbc/tools/pbc.py#L473

        """
        b = torch.inverse(a.transpose(-2, -1))
        heights_inv = torch.norm(b, dim=-1).detach().numpy()  # (ndim)
        nimgs = (rcut * heights_inv + 1.1).astype(np.int32)  # (ndim)

        n1_0 = torch.arange(-nimgs[0],
                            nimgs[0] + 1,
                            dtype=torch.int32,
                            device=self.device)  # (nimgs2,)
        n1_1 = torch.arange(-nimgs[1],
                            nimgs[1] + 1,
                            dtype=torch.int32,
                            device=self.device)  # (nimgs2,)
        n1_2 = torch.arange(-nimgs[2],
                            nimgs[2] + 1,
                            dtype=torch.int32,
                            device=self.device)  # (nimgs2,)
        ls = n1_2[:, None] * a[0, :]  # (nimgs2, ndim)
        ls = ls + n1_1[:, None, None] * a[1, :]  # (nimgs2, nimgs2, ndim)
        ls = ls + n1_0[:, None, None,
                       None] * a[2, :]  # (nimgs2, nimgs2, nimgs2, ndim)
        ls = ls.view(-1, ls.shape[-1])  # (nb, ndim)

        # drop ls that has norm > rcut * 1.05
        ls = ls[ls.norm(dim=-1) <= rcut * 1.05, :]  # (nb2, ndim)

        if exclude_zeros:
            ls = ls[torch.any(ls != 0, dim=-1), :]

        return ls



import numpy as np
import deepchem as dc
try:
    import pysam
except ImportError:
    print("Error: Unable to import pysam. Please make sure it is installed.")
from deepchem.feat import Featurizer


class SAMFeaturizer(Featurizer):
    """
    Featurizes SAM files, that store biological sequences aligned to a reference
    sequence. This class extracts Query Name, Query Sequence, Query Length,
    Reference Name,Reference Start, CIGAR and Mapping Quality of each read in
    a SAM file.

    Examples
    --------
    >>> from deepchem.data.data_loader import SAMLoader
    >>> import deepchem as dc
    >>> inputs = 'deepchem/data/tests/example.sam'
    >>> featurizer = dc.feat.SAMFeaturizer()
    >>> features = featurizer.featurize(inputs)
    Information for each read is stored in a 'numpy.ndarray'.
    >>> type(features[0])
    <class 'numpy.ndarray'>

    This is the default featurizer used by SAMLoader, and it extracts the following
    fields from each read in each SAM file in the given order:-
    - Column 0: Query Name
    - Column 1: Query Sequence
    - Column 2: Query Length
    - Column 3: Reference Name
    - Column 4: Reference Start
    - Column 5: CIGAR
    - Column 6: Mapping Quality
    For the given example, to extract specific features, we do the following.
    >>> features[0][0]     # Query Name
    r001
    >>> features[0][1]     # Query Sequence
    TTAGATAAAGAGGATACTG
    >>> features[0][2]     # Query Length
    19
    >>> features[0][3]     # Reference Name
    ref
    >>> features[0][4]     # Reference Start
    6
    >>> features[0][5]     # CIGAR
    [(0, 8), (1, 4), (0, 4), (2, 1), (0, 3)]
    >>> features[0][6]     # Mapping Quality
    30

    Note
    ----
    This class requires pysam to be installed. Pysam can be used with Linux or MacOS X.
    To use Pysam on Windows, use Windows Subsystem for Linux(WSL).
    """

    def __init__(self, max_records=None):
        """
        Initialize SAMFeaturizer.

        Parameters
        ----------
        max_records : int or None, optional
            The maximum number of records to extract from the SAM file. If None, all records will be extracted.
        """
        self.max_records = max_records

    def _featurize(self, datapoint):
        """
        Extract features from a SAM file.

        Parameters
        ----------
        datapoint : str
            Name of SAM file.

        Returns
        -------
        features : numpy.ndarray
        A 2D NumPy array representing the extracted features.
        Each row corresponds to a SAM record, and columns represent different features.
            - Column 0: Query Name
            - Column 1: Query Sequence
            - Column 2: Query Length
            - Column 3: Reference Name
            - Column 4: Reference Start
            - Column 5: CIGAR
            - Column 6: Mapping Quality
        """

        features = []
        record_count = 0

        for record in datapoint:
            feature_vector = [
                record.query_name,
                record.query_sequence,
                record.query_length,
                record.reference_name,
                record.reference_start,
                record.cigar,
                record.mapping_quality,
            ]

            features.append(feature_vector)
            record_count += 1

            # Break the loop if max_records is set
            if self.max_records is not None and record_count >= self.max_records:
                break

        datapoint.close()

        return np.array(features, dtype="object")



import torch
import torch.nn as nn
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.losses import L2Loss


class LCNNBlock(nn.Module):
    """
    The Lattice Convolution layer of LCNN

    The following class implements the lattice convolution function which is
    based on graph convolution networks where,
    [1] Each atom is represented as a node
    [2] Adjacent atom based on distance are considered as neighbors.

    Operations in Lattice Convolution:

    [1] In graph aggregation step- node features of neighbors are concatenated and
    into a linear layer. But since diffrent permutation of order of neighbors could
    be considered because of which , diffrent permutation of the lattice
    structure are considered in diffrent symmetrical angles (0 , 60 ,120 180 , 240 , 300 )

    [2] After the linear layer on each permutations, they are added up for each node and
    each node is transformed into a vector.

    Examples
    --------
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models.lcnn import LCNNBlock
    >>> from deepchem.feat.graph_data import GraphData
    >>> import numpy as np
    >>> nodes = np.array([0, 1, 2])
    >>> x = np.zeros((nodes.size, nodes.max()+1))
    >>> x[np.arange(nodes.size),nodes] = 1
    >>> v = np.array([ 0,0, 0,0, 1,1, 1,1, 2,2, 2,2 ])
    >>> u = np.array([ 1,2, 2,1, 2,0, 0,2, 1,0, 0,1 ])
    >>> graph = GraphData(node_features=x, edge_index=np.array([u, v]))
    >>> model = LCNNBlock(3*2, 3, 2)
    >>> G = graph.to_dgl_graph()
    >>> x = G.ndata.pop('x')
    >>> print(model(G, x).shape)
    torch.Size([3, 3])

    """

    def __init__(self,
                 input_feature: int,
                 output_feature: int = 19,
                 n_permutation_list: int = 6,
                 dropout: float = 0.2,
                 UseBN: bool = True):
        """
        Lattice Convolution Layer used in the main model

        Parameters
        ----------
        input_feature: int
            Dimenion of the concatenated input vector. Node_feature_size*number of neighbors
        output_feature: int, default 19
            Dimension of feature size of the convolution
        n_permutation_list: int, default 6
            Diffrent permutations taken along diffrent directions
        dropout: float
            p value for dropout between 0.0 to 1.0
        UseBN: bool
            To use batch normalisation
        """
        super(LCNNBlock, self).__init__()

        self.conv_weights = nn.Linear(input_feature, output_feature)
        self.batch_norm = nn.BatchNorm1d(output_feature)
        self.UseBN = UseBN
        self.activation = Shifted_softplus()
        self.dropout = Custom_dropout(dropout, n_permutation_list)
        self.permutation = n_permutation_list

    def reduce_func(self, nodes):
        number_of_sites = nodes.mailbox['m'].shape[0]
        return {
            'X_site':
                nodes.mailbox['m'].view(number_of_sites, self.permutation, -1)
        }

    def forward(self, G, node_feats):
        """
        Update node representations.

        Parameters
        ----------
        G: DGLGraph
            DGLGraph for a batch of graphs.
        node_feats: torch.Tensor
            The node features. The shape is `(N, Node_feature_size)`.

        Returns
        -------
        node_feats: torch.Tensor
            The updated node features. The shape is `(N, Node_feature_size)`.
        """
        try:
            import dgl.function as fn
        except:
            raise ImportError("This class requires DGL to be installed.")
        G = G.local_var()
        G.ndata['x'] = node_feats
        G.update_all(fn.copy_u('x', 'm'), self.reduce_func)
        X = self.conv_weights(G.ndata['X_site'])
        X = torch.stack([self.batch_norm(X_i) for X_i in X])
        node_feats = torch.stack([self.dropout(X_i).sum(axis=0) for X_i in X])
        return node_feats


class Atom_Wise_Convolution(nn.Module):
    """
    Performs self convolution to each node
    """

    def __init__(self,
                 input_feature: int,
                 output_feature: int,
                 dropout: float = 0.2,
                 UseBN: bool = True):
        """
        Parameters
        ----------
        input_feature: int
            Size of input feature size
        output_feature: int
            Size of output feature size
        dropout: float, defult 0.2
            p value for dropout between 0.0 to 1.0
        UseBN: bool
            Setting it to True will perform Batch Normalisation
        """
        super(Atom_Wise_Convolution, self).__init__()
        self.conv_weights = nn.Linear(input_feature, output_feature)
        self.batch_norm = nn.LayerNorm(output_feature)
        self.UseBN = UseBN
        self.activation = Shifted_softplus()
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, node_feats):
        """
        Update node representations.

        Parameters
        ----------
        node_feats: torch.Tensor
            The node features. The shape is `(N, Node_feature_size)`.

        Returns
        -------
        node_feats: torch.Tensor
            The updated node features. The shape is `(N, Node_feature_size)`.

        """

        node_feats = self.conv_weights(node_feats)
        if self.UseBN:
            node_feats = self.batch_norm(node_feats)

        node_feats = self.activation(node_feats)
        node_feats = self.dropout(node_feats)

        return node_feats


class Shifted_softplus(nn.Module):
    """
    Performs a Shifter softplus loss, which modifies with a value of log(2)
    """

    def __init__(self):
        super(Shifted_softplus, self).__init__()
        self.act = nn.Softplus()
        self.shift = nn.Parameter(torch.tensor([0.69310]), False)

    def forward(self, X):
        """
        Applies the Activation function

        Parameters
        ----------
        node_feats: torch.Tensor
            The node features.

        Returns
        -------
        node_feats: torch.Tensor
            The updated node features.

        """
        node_feats = self.act(X) - self.shift
        return node_feats


class Custom_dropout(nn.Module):
    """
    An implementation for few , Given a task perform a rowise sum of 2-d
    matrix , you get a zero out the contribution of few of rows in the matrix
    Given, X a 2-d matrix consisting of row vectors (1-d) x1 , x2 ,..xn.
    Sum = x1 + 0.x2 + .. + 0.xi + .. +xn
    """

    def __init__(self, dp_rate: float, n_permutation: int):
        """
        Parameters
        ----------
        dp_rate: float
            p value of dropout.
        """
        super(Custom_dropout, self).__init__()
        self.dropout = nn.Dropout(p=dp_rate)
        self.ones = nn.Parameter(torch.ones(n_permutation), requires_grad=False)

    def forward(self, layer):
        """
        Returns
        -------
        node_feats: torch.Tensor
            Updated tensor.
        """
        mask = self.dropout(self.ones).view(layer.shape[0],
                                            1).repeat(1, layer.shape[1])
        return mask * layer


class LCNN(nn.Module):
    """
    The Lattice Convolution Neural Network (LCNN)

    This model takes lattice representation of Adsorbate Surface to predict
    coverage effects taking into consideration the adjacent elements interaction
    energies.

    The model follows the following steps

    [1] It performs n lattice convolution operations. For more details look at the LCNNBlock class
    [2] Followed by Linear layer transforming into sitewise_n_feature
    [3] Transformation to scalar value for each node.
    [4] Average of properties per each element in a configuration

    Examples
    --------
    >>> import deepchem as dc
    >>> from pymatgen.core import Structure
    >>> import numpy as np
    >>> PRIMITIVE_CELL = {
    ...   "lattice": [[2.818528, 0.0, 0.0],
    ...               [-1.409264, 2.440917, 0.0],
    ...               [0.0, 0.0, 25.508255]],
    ...   "coords": [[0.66667, 0.33333, 0.090221],
    ...              [0.33333, 0.66667, 0.18043936],
    ...              [0.0, 0.0, 0.27065772],
    ...              [0.66667, 0.33333, 0.36087608],
    ...              [0.33333, 0.66667, 0.45109444],
    ...              [0.0, 0.0, 0.49656991]],
    ...   "species": ['H', 'H', 'H', 'H', 'H', 'He'],
    ...   "site_properties": {'SiteTypes': ['S1', 'S1', 'S1', 'S1', 'S1', 'A1']}
    ... }
    >>> PRIMITIVE_CELL_INF0 = {
    ...    "cutoff": np.around(6.00),
    ...    "structure": Structure(**PRIMITIVE_CELL),
    ...    "aos": ['1', '0', '2'],
    ...    "pbc": [True, True, False],
    ...    "ns": 1,
    ...    "na": 1
    ... }
    >>> DATA_POINT = {
    ...   "lattice": [[1.409264, -2.440917, 0.0],
    ...               [4.227792, 2.440917, 0.0],
    ...               [0.0, 0.0, 23.17559]],
    ...   "coords": [[0.0, 0.0, 0.099299],
    ...              [0.0, 0.33333, 0.198598],
    ...              [0.5, 0.16667, 0.297897],
    ...              [0.0, 0.0, 0.397196],
    ...              [0.0, 0.33333, 0.496495],
    ...              [0.5, 0.5, 0.099299],
    ...              [0.5, 0.83333, 0.198598],
    ...              [0.0, 0.66667, 0.297897],
    ...              [0.5, 0.5, 0.397196],
    ...              [0.5, 0.83333, 0.496495],
    ...              [0.0, 0.66667, 0.54654766],
    ...              [0.5, 0.16667, 0.54654766]],
    ...   "species": ['H', 'H', 'H', 'H', 'H', 'H',
    ...               'H', 'H', 'H', 'H', 'He', 'He'],
    ...   "site_properties": {
    ...     "SiteTypes": ['S1', 'S1', 'S1', 'S1', 'S1',
    ...                   'S1', 'S1', 'S1', 'S1', 'S1',
    ...                   'A1', 'A1'],
    ...     "oss": ['-1', '-1', '-1', '-1', '-1', '-1',
    ...             '-1', '-1', '-1', '-1', '0', '2']
    ...                   }
    ... }
    >>> featuriser = dc.feat.LCNNFeaturizer(**PRIMITIVE_CELL_INF0)
    >>> lcnn_feat = featuriser._featurize(Structure(**DATA_POINT)).to_dgl_graph()
    >>> print(type(lcnn_feat))
    <class 'dgl.heterograph.DGLGraph'>
    >>> model = LCNN()
    >>> out = model(lcnn_feat)
    >>> print(type(out))
    <class 'torch.Tensor'>


    Refrences
    -----------

    [1] Jonathan Lym,Geun Ho Gu, Yousung Jung , and Dionisios G. Vlachos
    "Lattice Convolutional Neural Network Modeling of Adsorbate Coverage
    Effects" The Journal of Physical Chemistry

    [2] https://forum.deepchem.io/t/lattice-convolutional-neural-network-modeling-of-adsorbate-coverage-effects/124

    Notes
    -----
    This class requires DGL and PyTorch to be installed.
    """

    def __init__(self,
                 n_occupancy: int = 3,
                 n_neighbor_sites: int = 19,
                 n_permutation: int = 6,
                 n_task: int = 1,
                 dropout_rate: float = 0.2,
                 n_conv: int = 2,
                 n_features: int = 19,
                 sitewise_n_feature: int = 25):
        """
        Parameters
        ----------

        n_occupancy: int, default 3
            number of possible occupancy
        n_neighbor_sites_list: int, default 19
            Number of neighbors of each site.
        n_permutation: int, default 6
            Diffrent permutations taken along diffrent directions.
        n_task: int, default 1
            Number of tasks
        dropout_rate: float, default 0.2
            p value for dropout between 0.0 to 1.0
        nconv: int, default 2
            number of convolutions performed
        n_feature: int, default 19
            number of feature for each site
        sitewise_n_feature: int, default 25
            number of features for atoms for site-wise activation

        """
        super(LCNN, self).__init__()

        modules = [LCNNBlock(n_occupancy * n_neighbor_sites, n_features)]
        for i in range(n_conv - 1):
            modules.append(
                LCNNBlock(n_features * n_neighbor_sites, n_features,
                          n_permutation))

        self.LCNN_blocks = nn.Sequential(*modules)
        self.Atom_wise_Conv = Atom_Wise_Convolution(n_features,
                                                    sitewise_n_feature)
        self.Atom_wise_Lin = nn.Linear(sitewise_n_feature, sitewise_n_feature)
        self.fc = nn.Linear(sitewise_n_feature, n_task)
        self.activation = Shifted_softplus()

    def forward(self, G):
        """
        Parameters
        ----------
        G: DGLGraph
            DGLGraph for a batch of graphs.

        Returns
        -------
        y: torch.Tensor
            A single scalar value

        """
        try:
            import dgl
        except:
            raise ImportError("This class requires DGL to be installed.")
        G = G.local_var()
        node_feats = G.ndata.pop('x')
        for conv in self.LCNN_blocks:
            node_feats = conv(G, node_feats)
        node_feats = self.Atom_wise_Conv(node_feats)
        node_feats = self.Atom_wise_Lin(node_feats)
        G.ndata['new'] = self.activation(node_feats)
        y = dgl.mean_nodes(G, 'new')
        y = self.fc(y)
        return y


class LCNNModel(TorchModel):
    """
    Lattice Convolutional Neural Network (LCNN).
    Here is a simple example of code that uses the LCNNModel with
    Platinum 2d Adsorption dataset.

    This model takes arbitrary configurations of Molecules on an adsorbate and predicts
    their formation energy. These formation energies are found using DFT calculations and
    LCNNModel is to automate that process. This model defines a crystal graph using the
    distance between atoms. The crystal graph is an undirected regular graph (equal neighbours)
    and different permutations of the neighbours are pre-computed using the LCNNFeaturizer.
    On each node for each permutation, the neighbour nodes are concatenated which are further operated.
    This model has only a node representation. Please confirm the detail algorithms from [1]_.

    Examples
    --------
    >>>
    >> import deepchem as dc
    >> from pymatgen.core import Structure
    >> import numpy as np
    >> from deepchem.feat import LCNNFeaturizer
    >> from deepchem.molnet import load_Platinum_Adsorption
    >> PRIMITIVE_CELL = {
    ..   "lattice": [[2.818528, 0.0, 0.0],
    ..               [-1.409264, 2.440917, 0.0],
    ..               [0.0, 0.0, 25.508255]],
    ..   "coords": [[0.66667, 0.33333, 0.090221],
    ..              [0.33333, 0.66667, 0.18043936],
    ..              [0.0, 0.0, 0.27065772],
    ..              [0.66667, 0.33333, 0.36087608],
    ..              [0.33333, 0.66667, 0.45109444],
    ..              [0.0, 0.0, 0.49656991]],
    ..   "species": ['H', 'H', 'H', 'H', 'H', 'He'],
    ..   "site_properties": {'SiteTypes': ['S1', 'S1', 'S1', 'S1', 'S1', 'A1']}
    .. }
    >> PRIMITIVE_CELL_INF0 = {
    ..    "cutoff": np.around(6.00),
    ..    "structure": Structure(**PRIMITIVE_CELL),
    ..    "aos": ['1', '0', '2'],
    ..    "pbc": [True, True, False],
    ..    "ns": 1,
    ..    "na": 1
    .. }
    >> tasks, datasets, transformers = load_Platinum_Adsorption(
    ..    featurizer= LCNNFeaturizer( **PRIMITIVE_CELL_INF0)
    .. )
    >> train, val, test = datasets
    >> model = LCNNModel(mode='regression',
    ..                   batch_size=8,
    ..                   learning_rate=0.001)
    >> model = LCNN()
    >> out = model(lcnn_feat)
    >> model.fit(train, nb_epoch=10)


    References
    ----------
    .. [1] Jonathan Lym and Geun Ho Gu, J. Phys. Chem. C 2019, 123, 18951−18959.

    Notes
    -----
    This class requires DGL and PyTorch to be installed.
    """

    def __init__(self,
                 n_occupancy: int = 3,
                 n_neighbor_sites_list: int = 19,
                 n_permutation_list: int = 6,
                 n_task: int = 1,
                 dropout_rate: float = 0.4,
                 n_conv: int = 2,
                 n_features: int = 44,
                 sitewise_n_feature: int = 25,
                 **kwargs):
        """
        This class accepts all the keyword arguments from TorchModel.

        Parameters
        ----------

        n_occupancy: int, default 3
            number of possible occupancy.
        n_neighbor_sites_list: int, default 19
            Number of neighbors of each site.
        n_permutation: int, default 6
            Diffrent permutations taken along diffrent directions.
        n_task: int, default 1
            Number of tasks.
        dropout_rate: float, default 0.4
            p value for dropout between 0.0 to 1.0
        nconv: int, default 2
            number of convolutions performed.
        n_feature: int, default 44
            number of feature for each site.
        sitewise_n_feature: int, default 25
            number of features for atoms for site-wise activation.
        kwargs: Dict
            This class accepts all the keyword arguments from TorchModel.
        """

        def init_weights(m):
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight)

        model = LCNN(n_occupancy, n_neighbor_sites_list, n_permutation_list,
                     n_task, dropout_rate, n_conv, n_features,
                     sitewise_n_feature)
        model.apply(init_weights)
        loss = L2Loss()
        output_types = ['prediction']
        super(LCNNModel, self).__init__(model,
                                        loss=loss,
                                        output_types=output_types,
                                        **kwargs)

    def _prepare_batch(self, batch):
        """
        Create batch data for LCNN.

        Parameters
        ----------
        batch: Tuple
            The tuple are `(inputs, labels, weights)`.

        Returns
        -------
        inputs: DGLGraph
            DGLGraph for a batch of graphs.
        labels: List[torch.Tensor] or None
            The labels converted to torch.Tensor
        weights: List[torch.Tensor] or None
            The weights for each sample or sample/task pair converted to torch.Tensor
        """
        try:
            import dgl
        except:
            raise ImportError("This class requires DGL to be installed.")

        inputs, labels, weights = batch
        dgl_graphs = [graph.to_dgl_graph() for graph in inputs[0]]
        inputs = dgl.batch(dgl_graphs).to(self.device)
        _, labels, weights = super(LCNNModel, self)._prepare_batch(
            ([], labels, weights))
        return inputs, labels, weights



"""
Loads synthetic reaction datasets from USPTO.

This file contains loaders for synthetic reaction datasets from the US Patent Office. http://nextmovesoftware.com/blog/2014/02/27/unleashing-over-a-million-reactions-into-the-wild/.
"""
import os
import logging
from typing import List, Optional, Tuple, Union

import deepchem as dc
from deepchem.data import Dataset
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader

try:
    from transformers import RobertaTokenizerFast
    from deepchem.feat.reaction_featurizer import RxnFeaturizer
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)

DEFAULT_DIR = dc.utils.data_utils.get_data_dir()

USPTO_MIT_URL = "https://deepchemdata.s3.us-west-1.amazonaws.com/datasets/USPTO_MIT.csv"
USPTO_STEREO_URL = "https://deepchemdata.s3.us-west-1.amazonaws.com/datasets/USPTO_STEREO.csv"
USPTO_50K_URL = "https://deepchemdata.s3.us-west-1.amazonaws.com/datasets/USPTO_50K.csv"
USPTO_FULL_URL = "https://deepchemdata.s3.us-west-1.amazonaws.com/datasets/USPTO_FULL.csv"

USPTO_TASK: List[str] = []


class _USPTOLoader(_MolnetLoader):

    def __init__(self, *args, subset: str, sep_reagent: bool, **kwargs):
        super(_USPTOLoader, self).__init__(*args, **kwargs)
        self.subset = subset
        self.sep_reagent = sep_reagent
        self.name = 'USPTO_' + subset

    def create_dataset(self) -> Dataset:
        if self.subset not in ['MIT', 'STEREO', '50K', 'FULL']:
            raise ValueError("Valid Subset names are MIT, STEREO and 50K.")

        if self.subset == 'MIT':
            dataset_url = USPTO_MIT_URL

        if self.subset == 'STEREO':
            dataset_url = USPTO_STEREO_URL

        if self.subset == '50K':
            dataset_url = USPTO_50K_URL

        if self.subset == 'FULL':
            dataset_url = USPTO_FULL_URL
            if self.splitter == 'SpecifiedSplitter':
                raise ValueError(
                    "There is no pre computed split for the full dataset, use a custom split instead!"
                )

        dataset_file = os.path.join(self.data_dir, self.name + '.csv')

        if not os.path.exists(dataset_file):
            logger.info("Downloading dataset...")
            dc.utils.data_utils.download_url(url=dataset_url,
                                             dest_dir=self.data_dir)
            logger.info("Dataset download complete.")

        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="reactions",
                                   featurizer=self.featurizer)

        return loader.create_dataset(dataset_file, shard_size=8192)


def load_uspto(
    featurizer: Union[dc.feat.Featurizer, str] = "RxnFeaturizer",
    splitter: Union[dc.splits.Splitter, str, None] = None,
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    subset: str = "MIT",
    sep_reagent: bool = True,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load USPTO Datasets.

    The USPTO dataset consists of over 1.8 Million organic chemical reactions
    extracted from US patents and patent applications. The dataset contains the
    reactions in the form of reaction SMILES, which have the general format:
    reactant>reagent>product.

    Molnet provides ability to load subsets of the USPTO dataset namely MIT,
    STEREO and 50K. The MIT dataset contains around 479K reactions, curated by
    jin et al. The STEREO dataset contains around 1 Million Reactions, it does
    not have duplicates and the reactions include stereochemical information.
    The 50K dataset contatins 50,000 reactions and is the benchmark for
    retrosynthesis predictions. The reactions are additionally classified into 10
    reaction classes. The canonicalized version of the dataset used by the loader
    is the same as that used by Somnath et. al.

    The loader uses the SpecifiedSplitter to use the same splits as specified
    by Schwaller et. al and Dai et. al. Custom splitters could also be used. There
    is a toggle in the loader to skip the source/target transformation needed for
    seq2seq tasks. There is an additional toggle to load the dataset with the
    reagents and reactants separated or mixed. This alters the entries in source
    by replacing the '>' with '.' , effectively loading them as an unified
    SMILES string.

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut. If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data. Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    subset: str (default 'MIT')
        Subset of dataset to download. 'FULL', 'MIT', 'STEREO', and '50K' are supported.
    sep_reagent: bool (default True)
        Toggle to load dataset with reactants and reagents either separated or mixed.
    skip_transform: bool (default True)
        Toggle to skip the source/target transformation.

    Returns
    -------
    tasks, datasets, transformers: tuple
        tasks : list
            Column names corresponding to machine learning target variables.
        datasets : tuple
            train, validation, test splits of data as
            ``deepchem.data.datasets.Dataset`` instances.
        transformers : list
            ``deepchem.trans.transformers.Transformer`` instances applied
            to dataset.

    References
    ----------
    .. [1] Lowe, D. Chemical reactions from US patents (1976-Sep2016)
        (Version 1). figshare (2017). https://doi.org/10.6084/m9.figshare.5104873.v1
    .. [2] Somnath, Vignesh Ram, et al. "Learning graph models for retrosynthesis
        prediction." arXiv preprint arXiv:2006.07038 (2020).
    .. [3] Schwaller, Philippe, et al. "Molecular transformer: a model for
        uncertainty-calibrated chemical reaction prediction."
        ACS central science 5.9 (2019): 1572-1583.
    .. [4] Dai, Hanjun, et al. "Retrosynthesis prediction with conditional
        graph logic network." arXiv preprint arXiv:2001.01408 (2020).
    """

    tokenizer = RobertaTokenizerFast.from_pretrained(
        "seyonec/PubChem10M_SMILES_BPE_450k")

    if featurizer == "plain":
        featurizer = dc.feat.DummyFeaturizer()
    else:
        featurizer = RxnFeaturizer(tokenizer, sep_reagent=sep_reagent)

    loader = _USPTOLoader(featurizer,
                          splitter,
                          transformers,
                          USPTO_TASK,
                          data_dir,
                          save_dir,
                          subset=subset,
                          sep_reagent=sep_reagent,
                          **kwargs)
    return loader.load_dataset(loader.name, reload)
